
// PathParser.java
// Nima Taheri, 2017
// https://gist.github.com/nimatrueway/cc1668c16096e2f2184275d8f780e3a5


/* PathParser adopted from react-native-svg renderer written for Android
https://raw.githubusercontent.com/react-native-community/react-native-svg/a958668e7f2753e7587eaf8a39ea46133809dc1a/android/src/main/java/com/horcrux/svg/PropHelper.java

  Used in ShapeOps.java to extract all the paths from the text 
  extracted from a loaded SVG file
  by repeatedly calling PathParser.getPath()
*/


import java.awt.geom.*;
import java.util.*;
import java.util.regex.*;



public class PathParser
{
  private static final Pattern PATH_REG_EXP = Pattern.compile(
                   "[a-df-z]|[\\-+]?(?:[\\d.]e[\\-+]?|[^\\s\\-+,a-z])+",
                   Pattern.CASE_INSENSITIVE);

  private static final Pattern DECIMAL_REG_EXP = 
                            Pattern.compile("(\\.\\d+)(?=-?\\.)");

  private Matcher matcher;
  private Path2D path;
  private final String mString;

  private double xPen = 0f;
  private double yPen = 0f;
  private double xPenDown, yPenDown;
  private double xPivot = 0f;
  private double yPivot = 0f;

  private double scale = 1f;
  private boolean isValid = true;
  private boolean isPenDownSet = false;

  private String lastCmd, mLastValue;

  private LinkedList<Point2D> bezierCurves;
  private Point2D lastStartPt;



  public PathParser(String s, double scale)
  {
    this.scale = scale;
    mString = s;
  }



  public Path2D getPath()
  {
    path = new Path2D.Double();
    bezierCurves = new LinkedList<Point2D>();
    matcher = PATH_REG_EXP.matcher(DECIMAL_REG_EXP.
                       matcher(mString).replaceAll("$1,"));

    while (matcher.find() && isValid)
      execCmd(matcher.group());
    return path;
  } /* getPath */



  private void execCmd(String cmd)
  {
    switch (cmd) {

    // moveTo cmd
    case "m":
      move(getNextDouble(), getNextDouble());
      break;
    case "M":
      moveTo(getNextDouble(), getNextDouble());
      break;

    // lineTo cmd
    case "l":
      line(getNextDouble(), getNextDouble());
      break;
    case "L":
      lineTo(getNextDouble(), getNextDouble());
      break;

    // horizontalTo cmd
    case "h":
      line(getNextDouble(), 0);
      break;
    case "H":
      lineTo(getNextDouble(), yPen);
      break;

    // verticalTo cmd
    case "v":
      line(0, getNextDouble());
      break;
    case "V":
      lineTo(xPen, getNextDouble());
      break;

    // curveTo cmd
    case "c":
      curve(getNextDouble(), getNextDouble(), getNextDouble(),
            getNextDouble(), getNextDouble(), getNextDouble());
      break;
    case "C":
      curveTo(getNextDouble(), getNextDouble(), getNextDouble(),
              getNextDouble(), getNextDouble(), getNextDouble());
      break;

    // smoothCurveTo cmd
    case "s":
      smoothCurve(getNextDouble(), getNextDouble(),
                  getNextDouble(), getNextDouble());
      break;
    case "S":
      smoothCurveTo(getNextDouble(), getNextDouble(),
                    getNextDouble(), getNextDouble());
      break;

    // quadraticBezierCurveTo cmd
    case "q":
      quadraticBezierCurve(getNextDouble(), getNextDouble(),
                           getNextDouble(), getNextDouble());
      break;
    case "Q":
      quadraticBezierCurveTo(getNextDouble(), getNextDouble(),
                             getNextDouble(), getNextDouble());
      break;

    // smoothQuadraticBezierCurveTo cmd
    case "t":
      smoothQuadraticBezierCurve(getNextDouble(), getNextDouble());
      break;
    case "T":
      smoothQuadraticBezierCurveTo(getNextDouble(), getNextDouble());
      break;

    // arcTo cmd
    case "a":
      arc(getNextDouble(), getNextDouble(), getNextDouble(),
          getNextBoolean(), getNextBoolean(),
          getNextDouble(), getNextDouble());
      break;
    case "A":
      arcTo(getNextDouble(), getNextDouble(), getNextDouble(),
            getNextBoolean(), getNextBoolean(),
            getNextDouble(), getNextDouble());
      break;

    // close cmd
    case "Z":
    case "z":
      close();
      break;

    default:
      mLastValue = cmd;
      execCmd(lastCmd);
      return;
    } /* switch */

    lastCmd = cmd;
    if (cmd.equals("m"))
      lastCmd = "l";
    else if (cmd.equals("M"))
      lastCmd = "L";
  } /* end of execCmd() */



  private Point2D getPointMap(double x, double y)
  {  return new Point2D.Double(x*scale, y*scale);  }


  private Point2D clonePointMap(Point2D map)
  {  return new Point2D.Double(map.getX(), map.getY());   }


  private boolean getNextBoolean()
  {
    if (matcher.find())
      return matcher.group().equals("1");
    else {
      isValid = false;
      path = new Path2D.Double();
      return false;
    }
  } 


  private double getNextDouble()
  {
    if (mLastValue != null) {
      String lastValue = mLastValue;
      mLastValue = null;
      return Double.parseDouble(lastValue);
    }
    else if (matcher.find())
      return Double.parseDouble(matcher.group());
    else {
      isValid = false;
      path = new Path2D.Double();
      return 0;
    }
  } /* getNextDouble */


  private void move(double x, double y)
  {  moveTo(x+xPen, y+yPen);   }


  private void moveTo(double x, double y)
  {
    xPivot = xPen = x;
    yPivot = yPen = y;
    path.moveTo(x*scale, y*scale);

    lastStartPt = getPointMap(x,y);
    bezierCurves.add( getPointMap(x, y));
  } /* moveTo */


  private void line(double x, double y)
  {  lineTo(x+xPen, y+yPen);   }


  private void lineTo(double x, double y)
  {
    setPenDown();
    xPivot = xPen = x;
    yPivot = yPen = y;
    path.lineTo(x*scale, y*scale);

    LinkedList<Point2D> pts = new LinkedList<>();
    pts.add(getPointMap(x, y));
    pts.add(getPointMap(x, y));
    pts.add(getPointMap(x, y));
    bezierCurves.addAll(pts);
  } /* lineTo */


  private void curve(double c1x, double c1y,                      
                     double c2x, double c2y, 
                     double ex,double ey)
  { curveTo(c1x+xPen, c1y+yPen,
            c2x+xPen, c2y+yPen,
            ex+xPen, ey+yPen);
  }


  private void curveTo(double c1x, double c1y,
                       double c2x, double c2y,
                       double ex, double ey)
  { xPivot = c2x;
    yPivot = c2y;
    cubicTo(c1x, c1y, c2x, c2y, ex, ey);
  }


  private void cubicTo(double c1x, double c1y,
                       double c2x, double c2y,
                       double ex, double ey)
  { setPenDown();
    xPen = ex;
    yPen = ey;
    path.curveTo(c1x*scale, c1y*scale,
                  c2x*scale, c2y*scale,
                  ex*scale, ey*scale);

    LinkedList<Point2D> pts = new LinkedList<>();
    pts.add(getPointMap(c1x, c1y));
    pts.add(getPointMap(c2x, c2y));
    pts.add(getPointMap(ex, ey));
    bezierCurves.addAll(pts);
  } /* cubicTo */



  private void smoothCurve(double c1x, double c1y, double ex, double ey)
  {  smoothCurveTo(c1x+xPen, c1y+yPen, ex+xPen, ey+yPen);   }


  private void smoothCurveTo(double c1x, double c1y, double ex, double ey)
  {
    double c2x = c1x;
    double c2y = c1y;
    c1x = (xPen*2) - xPivot;
    c1y = (yPen*2) - yPivot;
    xPivot = c2x;
    yPivot = c2y;
    cubicTo(c1x, c1y, c2x, c2y, ex, ey);
  } /* smoothCurveTo */



  private void quadraticBezierCurve(double c1x, double c1y,
                                    double c2x, double c2y)
  { quadraticBezierCurveTo(c1x+xPen, c1y+yPen, 
                           c2x+xPen, c2y+yPen);
  }



  private void quadraticBezierCurveTo(double c1x, double c1y,
                                      double c2x, double c2y)
  { xPivot = c1x;
    yPivot = c1y;
    double ex = c2x;
    double ey = c2y;
    c2x = (ex+c1x*2)/3;
    c2y = (ey+c1y*2)/3;
    c1x = (xPen+c1x*2)/3;
    c1y = (yPen+c1y*2)/3;
    cubicTo(c1x, c1y, c2x, c2y, ex, ey);
  } 


  private void smoothQuadraticBezierCurve(double c1x, double c1y)
  {  smoothQuadraticBezierCurveTo(c1x+xPen, c1y+yPen);   }


  private void smoothQuadraticBezierCurveTo(double c1x, double c1y)
  {
    double c2x = c1x;
    double c2y = c1y;
    c1x = (xPen*2) - xPivot;
    c1y = (yPen*2) - yPivot;
    quadraticBezierCurveTo(c1x, c1y, c2x, c2y);
  } 



  private void arc(double rx, double ry, double rotation,
                   boolean isOuter, boolean isClockwise,
                   double x, double y)
  {  arcTo(rx, ry, rotation, isOuter, isClockwise, x+xPen, y+yPen);   }



  private void arcTo(double rx, double ry, double rotation,
                     boolean isOuter, boolean isClockwise,
                     double x, double y)
  { double tX = xPen;
    double tY = yPen;

    ry = Math.abs(ry == 0 ? (rx == 0 ? (y - tY) : rx) : ry);
    rx = Math.abs(rx == 0 ? (x - tX) : rx);

    if ((rx == 0) || (ry == 0) || ((x == tX) && (y == tY)) ) {
      lineTo(x, y);
      return;
    }

    double rad = Math.toRadians(rotation);
    double cos = Math.cos(rad);
    double sin = Math.sin(rad);
    x -= tX;
    y -= tY;

    // Ellipse Center
    double cx = cos*x/2+sin*y/2;
    double cy = -sin*x/2+cos*y/2;
    double rxry = rx*rx*ry*ry;
    double rycx = ry*ry*cx*cx;
    double rxcy = rx*rx*cy*cy;
    double a = rxry - rxcy - rycx;

    if (a < 0) {
      a = Math.sqrt(1 - a/rxry);
      rx *= a;
      ry *= a;
      cx = x/2;
      cy = y/2;
    }
    else {
      a = Math.sqrt(a/(rxcy+rycx));

      if (isOuter == isClockwise)
        a = -a;

      double cxd = -a*cy*rx/ry;
      double cyd =  a*cx*ry/rx;
      cx = cos*cxd - sin*cyd+x/2;
      cy = sin*cxd+cos*cyd+y/2;
    }

    // Rotation+Scale Transform
    double xx =  cos/rx;
    double yx = sin/rx;
    double xy = -sin/ry;
    double yy = cos/ry;

    // Start and End Angle
    double sa =  Math.atan2(xy*-cx+yy*-cy, xx*-cx+yx*-cy);
    double ea =  Math.atan2(xy*(x - cx)+yy*(y - cy),
                            xx*(x - cx)+yx*(y - cy));

    cx += tX;
    cy += tY;
    x += tX;
    y += tY;

    setPenDown();

    xPen = xPivot = x;
    yPen = yPivot = y;

    arcToBezier(cx, cy, rx, ry, sa, ea, isClockwise, rad);
  } /* arcTo */



  private void close()
  {
    if (isPenDownSet) {
      xPen = xPenDown;
      yPen = yPenDown;
      isPenDownSet = false;
      path.closePath();

      LinkedList<Point2D> pts = new LinkedList<>();
      pts.add( clonePointMap(lastStartPt));
      pts.add( clonePointMap(lastStartPt));
      pts.add( clonePointMap(lastStartPt));
      bezierCurves.addAll(pts);
    }
  } /* close */


  private void arcToBezier(double cx, double cy,
                           double rx, double ry,
                           double sa, double ea,
                           boolean isClockwise, double rad)
  {
    // Inverse Rotation+Scale Transform
    double cos =  Math.cos(rad);
    double sin =  Math.sin(rad);
    double xx = cos*rx;
    double yx = -sin*ry;
    double xy = sin*rx;
    double yy =  cos*ry;

    // Bezier Curve Approximation
    double arc = ea - sa;
    if ((arc < 0) && isClockwise)
      arc += Math.PI*2;
    else if ((arc > 0) && !isClockwise)
      arc -= Math.PI*2;

    int n = (int) Math.ceil(Math.abs(arc/(Math.PI/2)));

    double step = arc/n;
    double k = (4/3)* Math.tan(step/4);

    double x =  Math.cos(sa);
    double y =  Math.sin(sa);

    for (int i = 0; i < n; i++) {
      double cp1x = x - k*y;
      double cp1y = y+k*x;

      sa += step;
      x =  Math.cos(sa);
      y =  Math.sin(sa);

      double cp2x = x+k*y;
      double cp2y = y - k*x;

      path.curveTo(
        (cx + xx*cp1x + yx*cp1y)*scale,
        (cy + xy*cp1x + yy*cp1y)*scale,
        (cx + xx*cp2x + yx*cp2y)*scale,
        (cy + xy*cp2x + yy*cp2y)*scale,
        (cx + xx*x + yx*y)*scale,
        (cy + xy*x + yy*y)*scale
           );
    }
  } /* arcToBezier */


  private void setPenDown()
  {
    if (!isPenDownSet) {
      xPenDown = xPen;
      yPenDown = yPen;
      isPenDownSet = true;
    }
  } /* setPenDown */


}  // end of PathParse class