
// Pics.java
// Andrew Davison, May 2022, ad@fivedots.coe.psu.ac.th

/* image operations
*/

import java.io.*;
import javax.swing.*;
import java.util.*;

import java.awt.*;
import java.awt.geom.*;
import java.awt.event.*;
import java.awt.image.*;
import java.awt.color.*;

import javax.imageio.*;
import javax.imageio.metadata.*;
import javax.imageio.stream.*;

import javax.xml.transform.*;
import javax.xml.transform.dom.*;
import javax.xml.transform.stream.*;
import org.w3c.dom.*;




public class Pics
{
  // image flipping (used in flip())
  public static final int VERT = 0;
  public static final int HORIZ = 1;
  public static final int BOTH = 2;   // flip horizontally and vertically

  // drawing positions (used in draw())
  public static final int NORTH = 0;
  public static final int EAST = 1;
  public static final int SOUTH = 2;
  public static final int WEST = 3;
  public static final int NE = 4;
  public static final int SE = 5;
  public static final int SW = 6;
  public static final int NW = 7;
  public static final int CENTER = 8;

  public static final double MAX_COLOR_DIST = 100.0;



  public static BufferedImage load(String fnm)
  {
    try {
      BufferedImage im = ImageIO.read(new File(fnm));
      // reportTransparency(fnm, im);
      return im;
    }
    catch(IOException e) {
      System.out.println("Cannot load " + fnm + ":\n" + e);
      return null;
    }
  } // end of load()


  public static void save(BufferedImage im, String fnm)
  {
    try {
      File f = new File(fnm);
      ImageIO.write(im, "png", f);
      System.out.println("Saved image to " + fnm);
    }
    catch (IOException e) {
      System.out.println("Unable to save image to " + fnm);
    }
  }  // end of save()




  public static String getName(String pathNm)
  // get filename of path without the extension
  {
    try {
      File f = new File(pathNm);
      String fnm = f.getName();
      int lastIndex = fnm.lastIndexOf('.');
      if (lastIndex != -1)
        fnm = fnm.substring(0, lastIndex);
      return fnm;
    } 
    catch (Exception e) {
      System.out.println("Could not extract filename: " + e);
      return null;
    }
  }  // end of getName()


  public static String getExtension(String pathNm)
  // get extension of a filename in lowercase
  {
    try {
      File f = new File(pathNm);
      String fnm = f.getName();
      int lastIndex = fnm.lastIndexOf('.');
      if (lastIndex != -1)
        return fnm.substring(lastIndex+1).toLowerCase();
    } 
    catch (Exception e) {
      System.out.println("Could not extract filename extension: " + e);
    }
    return null;
  }  // end of getExtension()




  public static void assign(BufferedImage toIm,
                            BufferedImage fromIm)
  {
    Graphics2D g2d = createGraphics(toIm);
    g2d.drawImage(fromIm, 0, 0, null);
    g2d.dispose();
  }


  public static BufferedImage copy(BufferedImage im)
  {
    if (im == null) {
      System.out.println("copy: image is null");
      return null;
    }

    BufferedImage canvas = makeCanvas(im, im.getWidth(), im.getHeight());
    Graphics2D g2d = createGraphics(canvas);

    g2d.drawImage(im, 0, 0, null);
    g2d.dispose();
    return canvas;
  } // end of copy()



  public static BufferedImage setBorders(BufferedImage im, int w, int h)
  // draw image on a canvas of size (w x h), at the center
  {
    if (im == null) {
      System.out.println("copy: image is null");
      return null;
    }

    BufferedImage canvas = makeCanvas(im, w, h);
    Graphics2D g2d = createGraphics(canvas);

    int x = w/2 - im.getWidth()/2;
    int y = h/2 - im.getHeight()/2;
    g2d.drawImage(im, x, y, im.getWidth(), im.getHeight(),  null);
    g2d.dispose();
    return canvas;
  } // end of setBorders()




  private static BufferedImage makeCanvas(BufferedImage im, int w, int h)
  // create an empty canvas of size (w, h) and same transpareny as im
  {
    GraphicsEnvironment ge =
              GraphicsEnvironment.getLocalGraphicsEnvironment();
    GraphicsConfiguration gc =
              ge.getDefaultScreenDevice().getDefaultConfiguration();

    return  gc.createCompatibleImage(w, h,
                   im.getColorModel().getTransparency());
  } // end of makeCanvas()



  public static Graphics2D createGraphics(BufferedImage im)
  {
    Graphics2D g2d = im.createGraphics();

    // use antialiasing
    g2d.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                          RenderingHints.VALUE_ANTIALIAS_ON);

    // smoother (and slower) image transformations  (e.g. for resizing)
    g2d.setRenderingHint(RenderingHints.KEY_INTERPOLATION,
                          RenderingHints.VALUE_INTERPOLATION_BILINEAR);

    g2d.setRenderingHint(RenderingHints.KEY_RENDERING,
                           RenderingHints.VALUE_RENDER_QUALITY);

    g2d.setPaint(Color.BLACK);

    return g2d;
  }


  // --------- resize -------------------

  public static BufferedImage resize(BufferedImage im, double scale)
  {  return resize(im, scale, scale);  }


  public static BufferedImage resize(BufferedImage im,
								double wScale, double hScale)
  {
    if (im == null) {
      System.out.println("resize image is null");
      return null;
    }

    if (wScale <= 0) {
      System.out.println("width scale cannot be <= 0");
      wScale = 1.0;
    }
    if (hScale <= 0) {
      System.out.println("height scale cannot be <= 0");
      hScale = 1.0;
    }

    int nWidth = (int) (im.getWidth() * wScale);
    int nHeight = (int) (im.getHeight() * hScale);

    BufferedImage canvas =  makeCanvas(im, nWidth, nHeight);
    Graphics2D g2d = createGraphics(canvas);
    // g2d.setComposite(AlphaComposite.Src);

    g2d.drawImage(im, 0, 0, nWidth, nHeight, null);
    g2d.dispose();
    return canvas;
  } // end of resize()



  public static BufferedImage resizeTo(BufferedImage im,
								int nWidth, int nHeight)
  {
    if (im == null) {
      System.out.println("resizeTo image is null");
      return null;
    }

    if (nWidth <= 0) {
      System.out.println("new width cannot be <= 0");
      nWidth = im.getWidth();
    }
    if (nHeight <= 0) {
      System.out.println("new height cannot be <= 0");
      nHeight = im.getHeight();
    }

    double wScale = ((double)nWidth)/im.getWidth();
    double hScale = ((double)nHeight)/im.getHeight();
    return resize(im, wScale, hScale);
  }  // end of resizeTo()



  public static BufferedImage resizeToWidth(BufferedImage im,
								            int nWidth)
  {
    if (im == null) {
      System.out.println("resizeTo image is null");
      return null;
    }

    if (nWidth <= 0) {
      System.out.println("new width cannot be <= 0");
      return im;
    }

    double scale = ((double)nWidth)/im.getWidth();
    return resize(im, scale);
  }  // end of resizeToWidth()


  public static BufferedImage resizeToHeight(BufferedImage im,
								            int nHeight)
  {
    if (im == null) {
      System.out.println("resizeTo image is null");
      return null;
    }

    if (nHeight <= 0) {
      System.out.println("new height cannot be <= 0");
      return im;
    }

    double scale = ((double)nHeight)/im.getHeight();
    return resize(im, scale);
  }  // end of resizeToHeight()


  // --------- flip -------------------


  public static BufferedImage flip(BufferedImage im, int flipKind)
  {
    if (im == null) {
      System.out.println("flip image is null");
      return null;
    }

    int w = im.getWidth();
    int h = im.getHeight();

    BufferedImage canvas =  makeCanvas(im, w, h);
    Graphics2D g2d = createGraphics(canvas);
    // g2d.setComposite(AlphaComposite.Src);

    // draw in the flipped image
    if (flipKind == VERT)   // vertical axis
      g2d.drawImage(im, w, 0,  0, h,         // dest rectangle
                        0, 0,  w, h, null);  // source rectangle
    else if (flipKind == HORIZ)    // horizontal axis
      g2d.drawImage(im, 0, h,  w, 0,
                        0, 0,  w, h, null);
    else    // assume BOTH (same as 180 rotation)
      g2d.drawImage(im, w, h,  0, 0,
                        0, 0,  w, h, null);

    g2d.dispose();
    return canvas;
  } // end of flip()


  // --------- rotate -------------------


  public static BufferedImage rotate(BufferedImage im, double angle)
  /* Create a new BufferedImage which is the input image, rotated
     angle degrees ccw.
  */
  {
    if (im == null) {
      System.out.println("rotate image is null");
      return null;
    }

    int diagonal = getDiagonal(im);
    BufferedImage canvas =  makeCanvas(im, diagonal, diagonal);
    Graphics2D g2d = createGraphics(canvas);

    AffineTransform origAT = g2d.getTransform(); // save original transform

    // rotate the coord. system of the canvas around its center ccw
    AffineTransform rot = new AffineTransform();
    rot.rotate( Math.toRadians(-angle), canvas.getWidth()/2,
                                       canvas.getHeight()/2);
    g2d.transform(rot);

    int x = canvas.getWidth()/2 - im.getWidth()/2;
    int y = canvas.getHeight()/2 - im.getHeight()/2;
    g2d.drawImage(im, x, y, null);   // copy in the image

    g2d.setTransform(origAT);    // restore original transform
    g2d.dispose();

    return canvas;
  } // end of rotate()



  public static int getDiagonal(BufferedImage im)
  {
    int w = im.getWidth();
    int h = im.getHeight();
    return (int)Math.round( Math.sqrt(w*w + h*h));
  }


  // --------- transparency ops -------------------


  public static boolean hasAlpha(BufferedImage im)
  // does im have an alpha channel?
  {
    if (im == null)
      return false;
    return im.getColorModel().hasAlpha();
  }


  public static double transparencyPercent(BufferedImage im)
  {
    if (im == null)
      return 1.0;

    int w = im.getWidth();
    int h = im.getHeight();

    int[] argb = im.getRGB(0, 0, w, h, null, 0, w);
    int count = 0;
    for (int i=0; i < argb.length; i++)
      if ((argb[i]>>24) == 0x00)
        count++;

    return ((double)count)/(w*h);
  }



  public static boolean containsTransparency(BufferedImage im)
  {
    for (int i = 0; i < im.getHeight(); i++) {
      for (int j = 0; j < im.getWidth(); j++) {
        if (isTransparent(im, j, i))
          return true;
      }
    }
    return false;
  }

  public static boolean isTransparent(BufferedImage im, double x, double y)
  {  return isTransparent(im, (int)Math.round(x), (int)Math.round(y));  }


  public static boolean isTransparent(BufferedImage im, int x, int y)
  {
    if (im == null)
      return false;
    if ((x < 0) || (x > im.getWidth()-1) ||
        (y < 0) || (y > im.getHeight()-1))
      return true;   // false

    int pixel = im.getRGB(x,y);
    boolean isTrans = ((pixel>>24) == 0x00);
   // if (!isTrans)
   //   System.out.println("pixel: " + String.format("0x%08X",pixel));
    return isTrans;
  }


  public static BufferedImage alpha(BufferedImage im)
  {  return alpha(im, 0, 0);  }

  public static BufferedImage alpha(BufferedImage im, 
                                    int width, int height)
  /* Make a copy of im, but in a BufferedImage object
     with an alpha channel of size (width x height).
     The new dimension must be greater than the
     image's dimension.
 */
  {
    if (im == null) {
      System.out.println("image is null");
      return null;
    }
    int imWidth = im.getWidth();
    int imHeight = im.getHeight();

    int nWidth = imWidth;
    if (width > imWidth) {
      System.out.println("Adjusting image width to " + width);
      nWidth = width;
    }
    else if (width != 0)
      System.out.println("Specified new width is too small");

    int nHeight = imHeight;
    if (height > imHeight) {
      System.out.println("Adjusting image height to " + height);
      nHeight = height;
    }
    else if (height != 0)
      System.out.println("Specified new height is too small");
      

    BufferedImage canvas =
         new BufferedImage( nWidth, nHeight,
                            BufferedImage.TYPE_INT_ARGB);  // alpha channel
    Graphics2D g2d = createGraphics(canvas);

    // copy image
    int x = (nWidth - imWidth)/2;
    int y = (nHeight - imHeight)/2;
    g2d.drawImage(im, x, y, null);
    g2d.dispose();

    return canvas;
  } // end of alpha()



  public static void colorModel(BufferedImage im)
  {
    int transparency = im.getColorModel().getTransparency();

    System.out.print("Color model: ");
    switch(transparency) {
      case Transparency.OPAQUE:
        System.out.println("opaque");
        break;
      case Transparency.BITMASK:
        System.out.println("bitmask");
        break;
      case Transparency.TRANSLUCENT:
        System.out.println("translucent");
        break;
      default:
        System.out.println("unknown");
        break;
    }
  }  // end of reportTransparency()




  public static void showMetadata(String fnm) 
  {
    try {
      ImageInputStream iis = ImageIO.createImageInputStream(new File(fnm));
      Iterator<ImageReader> readers = ImageIO.getImageReaders(iis);
      if (readers.hasNext()) {
        ImageReader reader = readers.next();
        reader.setInput(iis, true);
	  
        // read metadata of first image
        IIOMetadata metadata = reader.getImageMetadata(0);
	 
        String[] names = metadata.getMetadataFormatNames();
        for (int i = 0; i < names.length; i++) {
          System.out.println("\nFormat name: " + names[i]);
          prettyPrintXML(metadata.getAsTree(names[i]));
        }
      }
    }
    catch (Exception e) 
    {  System.out.println(e);  }
  }  // end of showMetadata()



  private static final void prettyPrintXML(Node xml) 
  {
    try {
      StreamResult result = new StreamResult(new StringWriter());

      Transformer tf = TransformerFactory.newInstance().newTransformer();
      tf.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
      tf.setOutputProperty(OutputKeys.ENCODING, "UTF-8");
      tf.setOutputProperty(OutputKeys.INDENT, "yes");
      tf.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "2");

      tf.transform(new DOMSource(xml), result);

      String xmlString = result.getWriter().toString();
      System.out.println(xmlString);
    }
    catch (Exception e) 
    {  System.out.println(e);  }
  }  // end of prettyPrintXML()



  // ----------- color related ------------------


  public static BufferedImage colorChg(BufferedImage im,
                                     Color currCol, Color newCol)
  {  return colorChg(im, currCol, newCol, MAX_COLOR_DIST);  }


  public static BufferedImage colorChg(BufferedImage im,
                        Color currCol, Color newCol, double maxDist)
  // Change all pixels 'close' to a color to the new color
  {
    if (im == null) {
      System.out.println("changeColor image is null");
      return null;
    }
    int w = im.getWidth();
    int h = im.getHeight();
    int[] rgb = im.getRGB(0, 0, w, h, null, 0, w);

    BufferedImage canvas =  makeCanvas(im, w, h);

    Color c;
    int nRGB = newCol.getRGB();
    for (int i = 0; i < rgb.length; i++) {
      c = new Color(rgb[i]);
      if (colorDistance(c, currCol) < maxDist)
        rgb[i] = nRGB;
    }

    canvas.setRGB(0, 0, w, h, rgb, 0, w);
    return canvas;
  }  // end of colorChg()


  public static Triplet<Integer,Integer,Integer> getPixelColor(
                                      BufferedImage im, int x, int y)
  {
    if (im == null) {
      System.out.println("getPixelColor image is null");
      return null;
    }

    int width = im.getWidth();
    int height = im.getHeight();
    if ((x < 0) || (x > width-1) || (y < 0) || (y > height-1)) {
      System.out.println("getPixelColor coordinate is out of range");
      return null;
    }

    int ARGB = im.getRGB(x, y);
    // int alpha = (ARGB >> 24) & 0xff;
    int red = (ARGB >> 16) & 0xff;
    int green = (ARGB >> 8) & 0xff;
    int blue = ARGB & 0xff;

    return new Triplet<Integer,Integer,Integer>(red, green, blue);

  }  // end of getPixelColor()


  public static boolean inside(BufferedImage im, int x, int y)
  // (x,y) is inside the dimensions of the image
  {
    int width = im.getWidth();
    int height = im.getHeight();
    return ((x >= 0) && (x < width) && (y >= 0) && (y < height));
  }



  public static TexturePaint loadTex(String fnm)
  {
    BufferedImage im = Pics.load(fnm);
    if (im == null)
      return null;

    Rectangle r = new Rectangle(0, 0, im.getWidth(), im.getHeight());
    return new TexturePaint(im, r);
  }


  public static double colorDistance(Color c1, Color c2)
  /* https://stackoverflow.com/questions/9018016/
       how-to-compare-two-colors-for-similarity-difference
     https://www.compuphase.com/cmetric.htm
  */
  {
    int r1 = c1.getRed();
    int r2 = c2.getRed();
    int redMean = (r1 + r2)/2;

    int r = r1 - r2;
    int g = c1.getGreen() - c2.getGreen();
    int b = c1.getBlue() - c2.getBlue();

    return Math.sqrt( (((512 + redMean)*r*r)>>8) + 4*g*g + 
                      (((767 - redMean)*b*b)>>8) );
  }  // end of colorDistance()



  // --------- blend two images -------------------


  public static BufferedImage blend(BufferedImage im1,
                                    BufferedImage im2)
  {  return blend(im1, im2, 0.5);  }


  public static BufferedImage blend(BufferedImage im1,
                                    BufferedImage im2, double weight)
  /* Blend the contents of two BufferedImages according to a specified
     weight which is the percentage of the first image to keep
  */
  {
    if (im1 == null) {
      System.out.println("First image is null");
      return im2;
    }

    if (im2 == null) {
      System.out.println("Second image is null");
      return im1;
    }

    // ensure the images are the same size
    int w1 = im1.getWidth();
    int w2 = im2.getWidth();
    int w = (w1 > w2) ? w1 : w2;

    int h1 = im1.getHeight();
    int h2 = im2.getHeight();
    int h = (h1 > h2) ? h1 : h2;

    BufferedImage b1 = setBorders(im1, w, h);
    BufferedImage b2 = setBorders(im2, w, h);

    BufferedImage blend =  new BufferedImage(w, h,
                                BufferedImage.TYPE_INT_ARGB);  // alpha channel
    Graphics2D g2d = createGraphics(blend);

    int x = (blend.getWidth() - im1.getWidth())/2;
    int y = (blend.getHeight() - im1.getHeight())/2;
    g2d.drawImage(im1, x, y, null);

   //g2d.setComposite(AlphaComposite.SrcOver.derive(0.5f));
    g2d.setComposite(AlphaComposite.getInstance(AlphaComposite.SRC_OVER,
                         (float) (1.0-weight)));
                    // 1.0f));

    x = (blend.getWidth() - im2.getWidth())/2;
    y = (blend.getHeight() - im2.getHeight())/2;
    g2d.drawImage(im2, x, y, null);
    g2d.dispose();

    return blend;
  }  // end of blend()



  // ---------------- drawing -----------------------------


  public static void draw(Graphics2D g2d, BufferedImage im,
                          int x, int y)
  {  draw(g2d, im, x, y, NW);  }


  public static void draw(Graphics2D g2d, BufferedImage im,
                          int x, int y, int pos)
  {
     if (im == null) {
       System.out.println("Null image supplied");
       return;
     }

     int w = im.getWidth();
     int h = im.getHeight();
     if (pos == NORTH)
       g2d.drawImage(im, x-w/2, y, null);
     else if (pos == EAST)
       g2d.drawImage(im, x-w, y-h/2, null);
     else if (pos == SOUTH)
       g2d.drawImage(im, x-w/2, y-h, null);
     else if (pos == WEST)
       g2d.drawImage(im, x, y-h/2, null);
     else if (pos == NE)
       g2d.drawImage(im, x-w, y, null);
     else if (pos == SE)
       g2d.drawImage(im, x-w, y-h, null);
     else if (pos == SW)
       g2d.drawImage(im, x, y-h, null);
     else if (pos == NW)
       g2d.drawImage(im, x, y, null);
     else if (pos == CENTER)
       g2d.drawImage(im, x-w/2, y-h/2, null);
     else {
       System.out.println("Unknown position: " + pos);
       g2d.drawImage(im, x, y, null);
     }
   } // end of draw()


  // --------- viewing window ----------------



  public static ImageViewer view(String title, int screenIdx)
  {  
    GraphicsConfiguration gc = screenConfig(screenIdx);
    return view(title, gc);
  }


  public static ImageViewer view(String title, GraphicsConfiguration gc)
  {  
    Rectangle bounds = screenArea(gc);
    
    ImageViewer pane = new ImageViewer(bounds.width, bounds.height);

    JFrame f = new JFrame(title, gc);
    f.getContentPane().add(pane, BorderLayout.CENTER);

    f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

    f.pack();
    f.setResizable(false);
    //f.setLocation(bounds.x, bounds.y);
    f.setBounds(bounds);
    f.setVisible(true);

    return pane;
  } // end of view()



  public static ImageViewer view(String title)
  {  
    Dimension d = Toolkit.getDefaultToolkit().getScreenSize();
    return view(title, d.width, d.height);
  }


  public static ImageViewer view(String title, int w, int h)
  {
    ImageViewer pane = new ImageViewer(w, h);

    JFrame f = new JFrame(title);
    f.getContentPane().add(pane, BorderLayout.CENTER);

    f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

    f.pack();
    f.setResizable(false);
    f.setVisible(true);

    return pane;
  } // end of view()


  public static ImageViewer view(String title, int x, int y,
                                      int w, int h)
  {
    ImageViewer pane = new ImageViewer(w, h);

    JFrame f = new JFrame(title);
    f.getContentPane().add(pane, BorderLayout.CENTER);

    f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

    f.setLocation(x, y);
    f.pack();
    f.setResizable(false);
    f.setVisible(true);

    return pane;
  } // end of view()


  // -------------- monitor ops --------------------



  public static GraphicsConfiguration altScreen() 
  {
	GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
    GraphicsDevice defScr = ge.getDefaultScreenDevice();

	GraphicsDevice[] gs = ge.getScreenDevices();
    for (int i = 0; i < gs.length; i++)
      if (!gs[i].equals(defScr))
        return gs[i].getDefaultConfiguration();

    System.out.println("No alternative screen found; using default");
    return defScr.getDefaultConfiguration();
  }  // end of altScreen()



  public static GraphicsConfiguration screenConfig(int screenIdx) 
  {
	GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
    GraphicsDevice defScr = ge.getDefaultScreenDevice();

	GraphicsDevice[] gs = ge.getScreenDevices();
    if ((screenIdx < 0) || (screenIdx >  gs.length-1)) {
      System.out.println("Screen index out of range; using default");
      return defScr.getDefaultConfiguration();
    }
    else 
      return gs[screenIdx].getDefaultConfiguration();
  }  // end of screenConfig()



  public static Rectangle screenArea(GraphicsConfiguration gc)
  {
    Rectangle bounds = gc.getBounds();

    Insets scrInsets = Toolkit.getDefaultToolkit().getScreenInsets(gc);
    
    Rectangle scrArea = new Rectangle();

    scrArea.x = bounds.x + scrInsets.left;
    scrArea.y = bounds.y + scrInsets.top;
    scrArea.height = bounds.height - (scrInsets.top + scrInsets.bottom);        
    scrArea.width = bounds.width - (scrInsets.left + scrInsets.right);

    return scrArea;
  }  // end of screenArea()




  // -------- other ops ---------------------

  public static void centerText(Graphics2D g2d, String s, int xc, int yc)
  // draw string s centered at (xc,yc)
  {
    FontMetrics fm = g2d.getFontMetrics();

    int x = xc - fm.stringWidth(s)/2;
    int y = yc - fm.getHeight()/2 + fm.getAscent();

    g2d.drawString(s, x, y);
  }


  public static void pause(int ms)
  {
    try {
      Thread.sleep(ms);
    }
    catch(InterruptedException ex) { }
  }  // end of pause()




  // ------------ test rig -----------

  public static void main(String args[])
  {
    ImageViewer iv = Pics.view("Tests");
    BufferedImage scrIm = iv.getImage();
    Graphics2D g2d = Pics.createGraphics(scrIm);


    BufferedImage fish1 = Pics.load("data/fish.png");
    System.out.println("Color at (60,90): " + Pics.getPixelColor(fish1, 60,90));

    BufferedImage fish2 = Pics.load("data/fish.png");

    BufferedImage im = Pics.resize(fish1, 0.5);
    // im = Pics.flip(im, Pics.VERT);
    // BufferedImage rotIm = Pics.rotate(im, 45);
    // im = Pics.blend(im, rotIm);

    BufferedImage colIm = Pics.colorChg(im, Color.WHITE, Color.GREEN);
    System.out.println("Transparency at (10,15): " + Pics.isTransparent(colIm, 10,15));

    g2d.drawImage(colIm, 0, 0, null);
    iv.repaint();
    Pics.pause(500);

    BufferedImage rotIm = Pics.rotate(im, 45);
    g2d.drawImage(rotIm, scrIm.getWidth()/2, 0, null);
    iv.repaint();

    Color nearWhite = new Color(242, 231, 230);
    System.out.println("Color dist to WHITE: " + 
           Pics.colorDistance(Color.WHITE, nearWhite));
    System.out.println("Color dist to BLACK: " + 
           Pics.colorDistance(Color.BLACK, nearWhite));

  } // end of main()



} // end of Pics