
// Search.java
// Andrew Davison, May 2022, ad@fivedots.coe.psu.ac.th

/* Two general search functions for placing tiles in an image.

   There is definitely room for improvement in the tiling quality
   of these functions in terms of the tile coverage (i.e. amount of
   space left at the end). 
*/

import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.util.*;
import java.util.function.*;



public class Search
{
  private static final int MAX_TILES = 600;
  private static final double ANGLE_EPS = 1.0;   // in degrees



  public static void tileLocs(Tile startTile, PointPairs ptPairs,
                           ImageViewer iview,
                           Function<String, Tile> selector)
  {  tileLocs(startTile, ptPairs, iview, selector, MAX_TILES);  }



  public static void tileLocs(Tile startTile,
                  PointPairs ptPairs,
                  ImageViewer iview,
                  Function<String, Tile> selector, int max)
  {
    startTile.draw( iview.getImage());

    ArrayList<Tile> placed = new ArrayList<>();
         // store drawn tiles in a list, and loop through it
         // note: a backtracking algorithm is not used
    placed.add(startTile);

     // look at every point of every drawn tile
    int tNum = 0;
    while((tNum < placed.size()) && (placed.size() < max)) {
      Tile t = placed.get(tNum);
      for (int i=0; i < t.numPts(); i++) {
        String ptName = t.ptName(i);
        double arcAngle = t.angle(ptName);
        boolean isDrawn = completeArc(arcAngle, t, ptName, placed,
                                      iview, ptPairs, max, selector);
            // try to add tiles so that this point is surrounded
        // note: failure/success of completeArc() isn't used to control the loop
      }
      tNum++;
    }

    System.out.println(placed.size() + " tiles placed");
    double tp = Pics.transparencyPercent(iview.getImage());
    System.out.printf("Unfilled space: %.2f%%\n", (tp*100));
                              // a simple coverage metric
  }  // end of tileLocs()




  private static boolean completeArc(double arcAngle, Tile t, String ptName,
                      ArrayList<Tile> placed, 
                      ImageViewer iview, PointPairs ptPairs,
                      int max, Function<String, Tile> selector)
  /* Add tiles so that the arcAngle increases to 360.
     This means that the location is completely surrounded
     by tiles. 
     This function uses backtracking while increasing the angle.
  */
  {
    if ((placed.size() == max) || (Math.abs(arcAngle-360.0) < ANGLE_EPS))
      return true;

    t.checkSides( iview.getImage());  // update the used sides for t
    String ptName1 = t.nextName(ptName);

    if ( iview.outsidePane(t.getPtLoc(ptName)) ||
         t.isUsedSide(ptName, ptName1) )
      return true;

    /* if we reach this line then we know that this tile point 
       is not outside the pane and the ptName--ptName1 side is unused,
       so we can start looking at its matching points
    */
    ArrayList<String> matchPts = ptPairs.fromMatches(ptName);
    Collections.shuffle(matchPts);  
        /* vary the choice order; this will produce different 
           tilings for different runs */

    for(String matchPt : matchPts) {
      BufferedImage im = iview.getImage();
      t.checkSides(im);

      // generate a new tile using one of the match point, and rotate it
      Tile nTile = selector.apply(matchPt);   // calls genTile()
      double rotAngle = nTile.edgeRotation(matchPt, t, ptName);
      Tile rotTile = nTile.rotate(rotAngle);
      // rotTile.addLabel(""+placed.size());   
          // draw a number in the rendered tile; useful for debugging

      // try to place the rotated tile on the screen image
      BufferedImage newIm =
          rotTile.tryDrawAt(im, matchPt, t.getPtLoc(ptName));

      if (newIm != null) {   // drawing was successful
        rotTile.checkSides(newIm);
        if (rotTile.checkSidesMatch(placed, ptPairs)) {          
          iview.update(newIm);    // update the viewer's image

          placed.add(rotTile);    // update state before recursing
          double nAngle = rotTile.angle(matchPt);
          arcAngle += nAngle;

          boolean isDrawn = completeArc(arcAngle, rotTile, matchPt, placed,
                                         iview, ptPairs, max, selector);

          if (isDrawn)  // later tiles completed the arc; move onto next tile
            return true;
          else {   /* later drawing led to failure, so undo/backtrack 
                      drawing of this tile and state changes */
            Tile temp = placed.get(placed.size()-1); 
            temp.unsetDrawn();
            placed.remove(temp);
            arcAngle -= nAngle;

            iview.update(im);  // reinstate old image in viewer
          }
        }
      }
    } // end of for-loop

    return false;  // none of the match points worked
  }  // end of completeArc()



  // --------------------------------------

  /* Unlike tileLocs() tileSpacey() doesn't require that a tile point 
     is completely surrounded by tiles. This means that more spaces 
     will appear in the final tiling.

     Nevertheless, the coding is quite similar.
  */


  public static void tileSpacey(Tile startTile, PointPairs ptPairs,
                           ImageViewer iview,
                           Function<String, Tile> selector)
  {  tileSpacey(startTile, ptPairs, iview, selector, MAX_TILES);  }



  public static void tileSpacey(Tile startTile, PointPairs ptPairs,
                           ImageViewer iview, 
                           Function<String, Tile> selector, int max)
  {
    startTile.draw(iview.getImage());

    ArrayList<Tile> placed = new ArrayList<>();
    placed.add(startTile);   // one drawn tile at this stage

     // look at every point of every drawn tile
     // this iteration through the placed list is very like tileLocs()
    int tNum = 0;
    while((tNum < placed.size()) && (placed.size() < max)) {
      Tile t = placed.get(tNum);
      for (int i=0; i < t.numPts(); i++) { 
        String ptName = t.ptName(i);
        placeTile(t, ptName, placed, iview, ptPairs, max, selector);
           // add another tile that matches the side for this point
      }
      tNum++;
    }

    System.out.println(placed.size() + " tiles placed");
    double tp = Pics.transparencyPercent(iview.getImage());
    System.out.printf("Unfilled space: %.2f%%\n", (tp*100));
  }  // end of tileSpacey()



  private static void placeTile(Tile t, String ptName,
              ArrayList<Tile> placed,
              ImageViewer iview, PointPairs ptPairs,
              int max, Function<String, Tile> selector)
  // draw another tile that matches the supplied point
  {
    if (placed.size() == max)
      return;

    t.checkSides( iview.getImage());  // update the used sides for t
    String ptName1 = t.nextName(ptName);

    if ( iview.outsidePane(t.getPtLoc(ptName)) ||
         t.isUsedSide(ptName, ptName1) )
      return;

    /* if we reach this line then we know that this tile point 
       is not outside the pane and the ptName--ptName1 side is unused,
       so we can start looking at its matching points
    */
    ArrayList<String> matchPts = ptPairs.fromMatches(ptName);
    Collections.shuffle(matchPts);  
        /* vary the choice order; this will produce different 
           tilings for different runs */

    // find a single matching point that lets a new tile be drawn
    boolean tileDrawn = false;
    int mNum = 0;
    while ((mNum < matchPts.size()) && !tileDrawn) {
      BufferedImage im = iview.getImage();
      t.checkSides(im);

      // generate a new tile using one of the match point, and rotate it
      String matchPt = matchPts.get(mNum);  // calls genTile()
      Tile nTile = selector.apply(matchPt);
      double rotAngle = nTile.edgeRotation(matchPt, t, ptName);
      Tile rotTile = nTile.rotate(rotAngle);
      // rotTile.addLabel(""+placed.size()); 
          // draw a number in the rendered tile; useful for debugging

      // try to place the rotated tile on the screen image
      BufferedImage newIm =
                rotTile.tryDrawAt(im, matchPt, t.getPtLoc(ptName));

      if (newIm != null) {  // drawing was successful
        rotTile.checkSides(newIm);
        if (rotTile.checkSidesMatch(placed, ptPairs)) {
          Pics.assign(im, newIm);   // change the contents of im
          iview.repaint();          // and repaint

          placed.add(rotTile);  // record the newly placed tile
          tileDrawn = true;
        }
      }
      mNum++;
    }
  }  // end of placeTile()



}  // end of Search