
// ShapeOps.java
// Andrew Davison, May 2022, ad@fivedots.coe.psu.ac.th

/*
   Operations for Shapes and Points
*/

import java.awt.*;
import java.awt.geom.*;
import java.util.*;
import java.awt.image.*;
import java.io.*;

import javax.xml.parsers.*;
import org.w3c.dom.*;
import org.xml.sax.*;

import java.nio.file.*;
import java.nio.charset.StandardCharsets;

// import org.jfree.graphics2d.svg.*;
          // for creation of SVG strings and saving
          // see saveSVGPath()


public class ShapeOps
{
  private static final int BOUNDARY = 10;   // for SVG path output


  public static Shape points(double[] xs, double[] ys)
  {
    if (xs.length != ys.length) {
      System.out.println("Arrays not of equal size");
      return null;
    }

    Path2D.Double path =
       new Path2D.Double(Path2D.Double.WIND_EVEN_ODD, xs.length);

    path.moveTo(xs[0], ys[0]);
    for (int i = 1; i < xs.length; i++)
        path.lineTo(xs[i], ys[i]);
    path.closePath();
    return path;
  }  // end of points()


  public static Shape points(ArrayList<Point2D.Double> pts)
  {  return points( pts.toArray(new Point2D.Double[0]));  }


  public static Shape points(Point2D.Double[] pts)
  {
    Path2D.Double path =
       new Path2D.Double(Path2D.Double.WIND_EVEN_ODD, pts.length);

    path.moveTo(pts[0].x, pts[0].y);
    for (int i = 1; i < pts.length; i++)
        path.lineTo(pts[i].x, pts[i].y);
    path.closePath();
    return path;
  }  // end of points()



  public static Shape nPolygon(double xc, double yc, 
                               double sideLength, int numSides)
  // generate in ccw order, with y-axis going down
  {
    double angle = Math.toRadians(((double)360.0)/numSides);
    double radius = (sideLength*0.5)/Math.sin(angle*0.5);
    return radPolygon(xc, yc, radius, numSides);
  }


  public static Shape radPolygon(double xc, double yc, 
                                 double radius, int numSides)
  {
    Path2D.Double path =
       new Path2D.Double(Path2D.Double.WIND_EVEN_ODD, numSides);
  
    // generate in ccw order, with y-axis going down
    for (int i = 0; i < numSides; i++) {
      double x = (xc + radius
                        * Math.cos(i * 2.0 * Math.PI / (double)numSides));
      double y = (yc - radius
                        * Math.sin(i * 2.0 * Math.PI / (double)numSides));
      // System.out.println("nPolygon: (" + x + ", " + y + ")");
  
      if (i == 0)
        path.moveTo(x, y);
      else
        path.lineTo(x, y);
    }
    path.closePath();
    return path;
  }  // end of nPolygon



  public static ArrayList<Point2D.Double> tilePoints(
                    double x, double y, ArrayList<Double> vals)
  {
    double[] vs = new double[vals.size()];
    for (int i=0; i < vs.length; i++)
      vs[i] = vals.get(i);
    return tilePoints(x, y, vs);
  }


  public static ArrayList<Point2D.Double> tilePoints(
                           double x, double y, double[] vals)
  /* vals[] format: ccw-angle-to-line1 length-line1 
                    ccw-angle-to-line2 length-line2 ... 
     The points are assumed to be in CCW order
     angles are in degrees;
     line1 begins at (x,y) at ccw angle to x-axis
  */
  {
    if (vals.length <= 2) {   // one or less sides
      System.out.println("Insufficient sides for a tile");
      return null;
    }

    ArrayList<Point2D.Double> pts = new ArrayList<>();
    Point2D.Double currPt = new Point2D.Double(x,y);
    pts.add(currPt);
    double totRot = 0;

    for (int i=0; i < vals.length; i = i+2) {
      totRot += vals[i]; 
      currPt = endPoint(currPt, vals[i+1], totRot);   
                         // pt, length, angle
      pts.add(currPt);
    }

    return pts;
  }  // end of tileShape()
                                     



  public static Shape moveTo(Shape shape, double tx, double ty)
  // move to (tx,ty) with respect to the center of the shape's
  // bounding box
  {
    AffineTransform af = new AffineTransform();
    Rectangle2D box = shape.getBounds2D();

    double cx = box.getCenterX();
    double cy = box.getCenterY();

    af.translate(tx-cx, ty-cy);

    return af.createTransformedShape(shape);
  }  // end of moveTo()



  public static Shape flipHoriz(Shape s)
  {  return resize(s, -1, 1);  }

  public static Shape flipVert(Shape s)
  {  return resize(s, 1, -1);  }



  public static Shape resize(Shape shape, double scale)
  {  return ShapeOps.resize(shape, scale, scale);  }


  public static Shape resize(Shape shape, double sx, double sy)
  // scale with respect to the center of its bounding box
  {
    AffineTransform af = new AffineTransform();
    Rectangle2D box = shape.getBounds2D();

    // double x = box.getX();
    // double y = box.getY();
    double cx = box.getCenterX();
    double cy = box.getCenterY();

    // Note: the transformations get applied IN REVERSE ORDER
    af.translate(cx, cy);
    af.scale(sx, sy);
    af.translate(-cx, -cy);

    return af.createTransformedShape(shape);
  } /* resize */



  public static Shape rotate(Shape s, double angleInDegrees)
  // rotate a shape around its center
  {
    AffineTransform af = new AffineTransform();
    Rectangle2D box = s.getBounds2D();

    double cx = box.getCenterX();
    double cy = box.getCenterY();
    // System.out.println("rotate center: (" + cx + ", " + cy + ")");

    // NOTE: The transformations get applied IN REVERSE ORDER
    af.translate(cx, cy);
    af.rotate(Math.toRadians(-angleInDegrees));
                  // so rotation ccw since y-axis is downwards
    af.translate(-cx, -cy);

    return af.createTransformedShape(s);
  } /* rotate */



  public static void print(Shape shape)
  {
    double[] coords = new double[6];
    PathIterator pi = shape.getPathIterator(null);  // new AffineTransform());

    while (!pi.isDone()) {
      switch (pi.currentSegment(coords)) {
        case PathIterator.SEG_MOVETO:
            System.out.printf("Move to x1=%.2f, y1=%.2f\n", coords[0], coords[1]);
            break;

        case PathIterator.SEG_LINETO:
            System.out.printf("Line to x1=%.2f, y1=%.2f\n", coords[0], coords[1]);
            break;

        case PathIterator.SEG_QUADTO:
            System.out.printf("quad to x1=%.2f, y1=%.2f, x2=%.2f, y2=%.2f\n",
                             coords[0], coords[1], coords[2], coords[3]);
            break;

        case PathIterator.SEG_CUBICTO:
            System.out.printf("cubic to x1=%.2f, y1=%.2f, x2=%.2f, y2=%.2f, x3=%.2f, y3=%.2f\n",
                    coords[0], coords[1], coords[2], coords[3], coords[4], coords[5]);
            break;

        case PathIterator.SEG_CLOSE:
            System.out.printf("close\n");
            break;
      }
      pi.next();
    }
  }  // end of print()


  public static Shape combine(ArrayList<Shape> shapes)
  {
    Path2D.Double path = new Path2D.Double();
    for(Shape s : shapes)
      path.append(s, false);
    return path;
  }


  public static ArrayList<Point2D.Double> getPoints(Shape shape)
  {
    double coords[] = new double[6];
    PathIterator pi = shape.getPathIterator(null, 0.1);  // flatness
    int numPaths = 0;
    ArrayList<Point2D.Double> pts = new ArrayList<>();

    while (!pi.isDone()) {
      switch (pi.currentSegment(coords)) {
        case PathIterator.SEG_LINETO:
          pts.add( new Point2D.Double( coords[0], coords[1]));
          break;

        case PathIterator.SEG_MOVETO:
          pts.add( new Point2D.Double( coords[0], coords[1]));
          numPaths++;
          break;

        case PathIterator.SEG_CLOSE:
          if (numPaths > 1)
            System.out.println("Multiple subpaths found");
          break;

        case PathIterator.SEG_QUADTO:
          System.out.println("SEG_QUADTO in flattening path iterator");
          break;

        case PathIterator.SEG_CUBICTO:
          System.out.println("SEG_CUBICTO in flattening path iterator");
          break;
      }
      pi.next();
    }
    return pts;
  }  // end of getPoints()



  public static int area(Shape shape)
  {
    ArrayList<Point2D.Double> pts = getPoints(shape);
    Point2D.Double currPt = pts.get(0);
    Point2D.Double nextPt;
    int sum = 0;
    for (int i=1; i < pts.size(); i++) {
      nextPt = pts.get((i+1)%pts.size());
      sum +=  (currPt.x * nextPt.y) - (currPt.y * nextPt.x);
      currPt = nextPt;
    }
    return Math.abs(sum/2);
  }  // end of area()



  public static Point2D.Double getCenter(Shape shape)
  {
    Rectangle box = shape.getBounds();
    double cx = box.getCenterX();
    double cy = box.getCenterY();
    return new Point2D.Double(cx,cy);
  }


  public static Point2D.Double getMidPoint(Shape shape, int i)
  /* return a midpoint between the ith and i+1 pt in the
     shape after it has been converted to points, or null.
  */
  {
    ArrayList<Point2D.Double> pts = getPoints(shape);
    if (pts == null) {
      System.out.println("No points extracted");
      return null;
    }
    if (pts.size() <= 1)  {
      System.out.println("Not enough points extracted");
      return null;
    }

    if ((i < 0) || (i > pts.size()-1)) {
      System.out.println("Point2D.Double index out of range");
      return null;
    }

    Point2D.Double pt1 = pts.get(i);
    Point2D.Double pt2 = pts.get( (i+1)%pts.size());
    return new Point2D.Double( (pt1.x+pt2.x)/2, (pt1.y+pt2.y)/2);
  }  // end of getMidPoint()



  public static ArrayList<Point2D.Double> genExtras(Shape shape, int addFactor)
  /* convert the shape into a list of points *and* add addFactor extra
     points between every pair of points (including the last and first)
     by interpolating between them.
  */
  {
    ArrayList<Point2D.Double> pts = getPoints(shape);
    if (pts == null) {
      System.out.println("No points extracted");
      return null;
    }
    if (pts.size() <= 1)  {
      System.out.println("Not enough points extracted");
      return null;
    }

    if (addFactor < 1) {
      System.out.println("The addFactor must be 1 or more");
      return pts;
    }

    Point2D.Double currPt = pts.get(0);
    Point2D.Double nextPt;
    ArrayList<Point2D.Double> extendedPts = new ArrayList<>();
    for (int i=0; i < pts.size(); i++) {
      nextPt = pts.get((i+1)%pts.size());
      double xOffset = (nextPt.x - currPt.x)/(addFactor+1);
      double yOffset = (nextPt.y - currPt.y)/(addFactor+1);

      extendedPts.add(currPt);
      for (int j=1; j < addFactor+1; j++)
        extendedPts.add( new Point2D.Double( currPt.x + (xOffset*j),
                                    currPt.y + (yOffset*j)) );
      currPt = nextPt;
    }
    return extendedPts;
  }  // end of genExtras()



  public static boolean doIntersect(Shape s1, Shape s2)
  {
    Area area1 = new Area(s1);
    area1.intersect(new Area(s2));  // changes area1
    return !area1.isEmpty();
  }


  public static double intersectSize(Shape s1, Shape s2)
  {
    Area area1 = new Area(s1);
    area1.intersect(new Area(s2));
    if (area1.isEmpty())
      return 0;
    else {
      Rectangle2D r = area1.getBounds2D();
      return r.getWidth() * r.getHeight();
    }
  }


  public static void draw(Graphics2D g2d, Shape shape)
  {
    g2d.setPaint(Color.WHITE);
    g2d.fill(shape);
    g2d.setPaint(Color.BLACK);
    g2d.setStroke(new BasicStroke(3.0f, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL));
    g2d.draw(shape);
  }  // end of draw()



  public static void drawAt(Graphics2D g2d, Shape shape, int x, int y)
  // draw shape centered at (x,y)
  {
    Rectangle2D box = shape.getBounds2D();

    double xOffset = x - box.getCenterX();
    double yOffset = y - box.getCenterY();

    AffineTransform af = new AffineTransform();
    af.translate(xOffset, yOffset);
    Shape movedShape = af.createTransformedShape(shape);

    g2d.setPaint(Color.WHITE);
    g2d.fill(movedShape);
    g2d.setPaint(Color.BLACK);
    g2d.setStroke(new BasicStroke(3.0f, BasicStroke.CAP_ROUND, BasicStroke.JOIN_BEVEL));
    g2d.draw(movedShape);
  }  // end of drawAt()


  // --------------- IO ops -----------------------



  public static ArrayList<Shape> readShapes(String fnm)
  {
    System.out.println("Reading from " + fnm);
    ArrayList<Shape> shapes = new ArrayList<>();
    try {
      Scanner sc = new Scanner( new File(fnm));
      while (sc.hasNext())
        shapes.add( readShape(sc));
      sc.close();
      return shapes;
    }
    catch(IOException e) {
      System.out.println("readShapes() error: " + e);
      return null;
    }
  }  // end of readShapes()



  private static Shape readShape(Scanner sc)
  {
    Shape result = null;
    try {
      String shapeName = sc.next();
      // System.out.println("Reading \"" + shapeName + "\"");
      if (shapeName.equals("Line2D")) {
        double x1 = sc.nextDouble();
        double y1 = sc.nextDouble();
        double x2 = sc.nextDouble();
        double y2 = sc.nextDouble();
        result = new Line2D.Double(x1, y1, x2, y2);
      }
      else if (shapeName.equals("Rectangle2D")) {
        double x = sc.nextDouble();
        double y = sc.nextDouble();
        double w = sc.nextDouble();
        double h = sc.nextDouble();
        result = new Rectangle2D.Double(x, y, w, h);
      }
      else if (shapeName.equals("Ellipse2D")) {
        double x = sc.nextDouble();
        double y = sc.nextDouble();
        double w = sc.nextDouble();
        double h = sc.nextDouble();
        result = new Ellipse2D.Double(x, y, w, h);
      }
      else if (shapeName.equals("Arc2D")) {
        double x = sc.nextDouble();
        double y = sc.nextDouble();
        double w = sc.nextDouble();
        double h = sc.nextDouble();
        double as = sc.nextDouble();     // Angle Start
        double ae = sc.nextDouble();     // Angle Extent
        int at = sc.nextInt();           // Arc type
        result = new Arc2D.Double(x, y, w, h, as, ae, at);
      }
      else if (shapeName.equals("Path2D")) {
        Path2D.Double gp = new Path2D.Double();
        float[] args = new float[6];
        boolean hasNext = sc.nextBoolean();
        while (!hasNext) {
          int type = sc.nextInt();
          for (int i = 0; i < 6; i++)
            args[i] = sc.nextFloat();
     
          switch (type) {
            case PathIterator.SEG_MOVETO:
              gp.moveTo(args[0], args[1]);
              break;
            case PathIterator.SEG_LINETO:
              gp.lineTo(args[0], args[1]);
              break;
            case PathIterator.SEG_CUBICTO:
              gp.curveTo(args[0], args[1], args[2],
                         args[3], args[4], args[5]);
              break;
            case PathIterator.SEG_QUADTO:
              gp.quadTo(args[0], args[1], args[2], args[3]);
              break;
            case PathIterator.SEG_CLOSE:
              gp.closePath();
              break;
            default:
              System.out.println("Unknown path type");
          } /* switch */
          gp.setWindingRule(sc.nextInt());
          hasNext = sc.nextBoolean();
        }
        result = gp;
      }  // end of "Path2D" branch
      else if (shapeName.equals("Tile")) {
        // x y <angle length>* $
        double x = sc.nextDouble();
        double y = sc.nextDouble();
        ArrayList<Double> vals = new ArrayList<>();
        String valStr = sc.next();
        while (!valStr.equals("$")) {
          vals.add( Double.parseDouble(valStr));  // angle
          vals.add( sc.nextDouble());    // length
          valStr = sc.next();
        }
        return points( tilePoints(x, y, vals));
      }  // end of "Tile" branch
      else
        System.out.println("Unknown shape name: " + shapeName);
      return result;
    }
    catch(Exception e) {
      System.out.println("readShape() error: " + e);
      return null;
    }
  }  // end of readShape()



  public static void writeShapes(String fnm, ArrayList<Shape> shapes)
  {
    System.out.println("Writing to " + fnm);
    try {
      PrintWriter pw = new PrintWriter( new FileWriter( new File(fnm)));
      for(Shape s : shapes)
        writeShape(s, pw);
      pw.close();
    }
    catch(IOException e) {
      System.out.println("writeShapes() error: " + e);
    }
  }  // end of writeShapes()


  private static void writeShape(Shape shape, PrintWriter pw)
  {
    try {
      if (shape instanceof Line2D) {
        Line2D line = (Line2D) shape;
        pw.print("Line2D ");
        pw.println(line.getX1() + " " + line.getY1() + " " +
                 line.getX2() + " " + line.getY2());
      }
      else if (shape instanceof Rectangle2D) {
        Rectangle2D rect = (Rectangle2D) shape;
        pw.print("Rectangle2D ");
        pw.print(rect.getX() + " " + rect.getY() + " ");
        pw.println(rect.getWidth() + " " + rect.getHeight());
      }
      else if (shape instanceof Ellipse2D) {
        Ellipse2D elp = (Ellipse2D) shape;
        pw.print("Ellipse2D ");
        pw.print(elp.getX() + " " + elp.getY() + " ");
        pw.println(elp.getWidth() + " " + elp.getHeight());
      }
      else if (shape instanceof Arc2D) {
        Arc2D arc = (Arc2D) shape;
        pw.print("Arc2D ");
        pw.print(arc.getX() + " " + arc.getY() + " ");
        pw.print(arc.getWidth() + " " + arc.getHeight() + " ");
        pw.print(arc.getAngleStart() + " ");
        pw.print(arc.getAngleExtent() + " ");
        pw.println(arc.getArcType());
      }
      else if (shape instanceof Path2D) {
        pw.println("Path2D ");
        PathIterator pi = shape.getPathIterator(null);
        float[] args = new float[6];
        pw.print("  " + pi.isDone() + " ");  // boolean
        while (!pi.isDone()) {
          int type = pi.currentSegment(args);
          pw.print(type + " ");
          // could only output the values
          // required for the segment type
          for (int i = 0; i < 6; i++)
            pw.print(args[i] + " ");
          pw.println(pi.getWindingRule() + " ");

          pi.next();
          pw.print("  " + pi.isDone() + " "); // boolean
        }
        pw.println();
      }
      else {
        System.out.println("Unrecognized shape: " + shape);
      }
    }
    catch(Exception e) {
      System.out.println("writeShape() error: " + e);
    }
  }  // end of writeShape()


  public static ArrayList<Shape> loadSVGPaths(String fnm)
  // load all the paths from a SVG file
  {
    ArrayList<Shape> paths = new ArrayList<>();
    ArrayList<String> pathStrs = loadSVGPathStrings(fnm);
    for(int i=0; i < pathStrs.size(); i++) {
      PathParser pp =  new PathParser(pathStrs.get(i), 1.0f);
      if (pp != null) {
        Shape s = pp.getPath();
        if (s == null)
          System.out.println("Unable to parse path " + i);
        else
          paths.add(s);
      }
      else
        System.out.println("Could not create parser for path " + i);
    }
    return paths;
  }  // end of loadSVGPaths()


  public static ArrayList<String> loadSVGPathStrings(String fnm)
  {
    ArrayList<String> pathStrs = new ArrayList<>();
    try {
      System.out.println("Parsing " + fnm);
      DocumentBuilder db = DocumentBuilderFactory.newInstance().
                                     newDocumentBuilder();
      Document doc = db.parse( new File(fnm));
      doc.getDocumentElement().normalize();
       
      // Print the root node
      Element root = doc.getDocumentElement();
      System.out.println("Root node is " + root.getNodeName());
       
      // Get all paths
      NodeList paths = doc.getElementsByTagName("path");
      System.out.println("No. of paths: " + paths.getLength());

      for(int i = 0; i < paths.getLength(); i++) {
        Node node = paths.item(i);
        // System.out.println(); 
        if (node.getNodeType() == Node.ELEMENT_NODE) {
          Element elem = (Element) node;
          String dStr = elem.getAttribute("d");
          // System.out.println((i+1) + ". d: \"" + dStr  + "\"");
          pathStrs.add(dStr);
        }
      }
    }
    catch(Exception e)
    {  System.out.println(e);  }

    return pathStrs;
  }  // end of loadSVGPathStrings()



  public static Shape loadPath(String fnm)
  // load a single SVG path from a text "path" file
  {
    try {
      System.out.println("Parsing " + fnm);      
      String pathStr = new String(
        Files.readAllBytes(Paths.get(fnm)), StandardCharsets.UTF_8);
      PathParser pp =  new PathParser(pathStr, 1.0f);
      if (pp != null) {
        Shape s = pp.getPath();
        if (s == null)
          System.out.println("Unable to parse path");
        else
          return s;
      }
      else
        System.out.println("Could not create parser for path");
    }
    catch(IOException e)
    {  System.out.println(e);  }

    return null;
  }  // end of loadPath()

/*
  public static void saveSVGPath(String fnm, Shape s)
  {
    Path2D.Double path = new Path2D.Double();
    path.append(s, false);
    savePath(fnm, path);
  }


  public static void saveSVGPath(String fnm, ArrayList<Shape> shapes)
  {
    Path2D.Double path = new Path2D.Double();
    for(Shape s : shapes)
      path.append(s, false);
    savePath(fnm, path);
  }


  private static void savePath(String fnm, Path2D.Double path)
  {
    Rectangle rect = path.getBounds();
    int scrWidth = (rect.x + rect.width) + 2*BOUNDARY;
    int scrHeight = (rect.y + rect.height) + 2*BOUNDARY;
     
    SVGGraphics2D svg2d = new SVGGraphics2D(scrWidth, scrHeight);
    svg2d.draw(path);

    String elem = svg2d.getSVGElement();
    // System.out.println("Path: " + elem);
    System.out.println("Saving SVG path to " + fnm);
    try {
      SVGUtils.writeToSVG(new File(fnm), elem);
    }
    catch(IOException e)
    {  System.out.println(e); }
  }  // end of saveSVGPath()
*/


  // --------------------- Point operations ----------------------


  public static Polar cart2Polar(Point2D.Double pt)
  {  return cart2Polar(pt.x, pt.y);  }


  public static Polar cart2Polar(double x, double y)
  // convert cartesian to Polar
  {
    double len = Math.sqrt(x*x + y*y);
    double theta = Math.atan2(y, x);  // in radians, -pi to pi
    double angle = Math.toDegrees(theta);
    if (angle < 0) // in 3rd or 4th quadrants
      angle = 360 + angle; // so in range 180 to 360

    return new Polar(angle, len);
  }


  public static Point2D.Double rotatePoint(Point2D.Double p1, Point2D.Double p2,
                                           double angleInDegrees)
  // Rotates p2 around p1 to get a new point
  {
     double dx = p2.getX() - p1.getX();
     double dy = p1.getY() - p2.getY();   // y-axis increase downwards

     double len = Math.hypot(dx, dy);
     double currAngle = Math.atan2(dy, dx);

     double ang = currAngle + Math.toRadians(angleInDegrees);
     double xLen = len * Math.cos(ang);
     double yLen = len * Math.sin(ang);

     return new Point2D.Double(p1.x + xLen, p1.y - yLen);
  }  // end of rotatePoint()




  public static Point2D.Double endPoint(Point2D.Double p1, double len,
                                        double angleInDegrees)
  // Calculate point at the end of len units from p1 at the specified angle
  {
     double ang = Math.toRadians(angleInDegrees);
     double xLen = len * Math.cos(ang);
     double yLen = len * Math.sin(ang);

     return new Point2D.Double(p1.x + xLen, p1.y - yLen);
  }  // end of endPoint()



  public static double pointAngle(Point2D.Double prevPt, Point2D.Double pt,
                                  Point2D.Double nextPt)
  // positive angle in degrees between prevPt, Pt, and nextPt in ccw order
  {
    double angle = Math.toDegrees(
                        Math.atan2(prevPt.y - pt.y, prevPt.x - pt.x) -
                        Math.atan2(nextPt.y - pt.y, nextPt.x - pt.x) );
    if (angle < 0)
      return -angle;
    else
      return 360 - angle;
  }  // end of pointAngle()



  public static double getAngle(Point2D.Double p1, Point2D.Double p2)
  {
    double angle = Math.toDegrees(Math.atan2(p2.y - p1.y, p2.x - p1.x));
    if (angle < 0)
      angle += 360;
    return angle;
  }  // end of getAngle()


  public static double getDistance(Point2D.Double p1, Point2D.Double p2)
  {  return Math.hypot(p1.x-p2.x, p1.y-p2.y);  }



  public static boolean inside(BufferedImage im, Point2D.Double pt)
  {  return inside(im, pt.x, pt.y);  }

  public static boolean inside(BufferedImage im, double x, double y)
  {  return Pics.inside(im, (int)Math.round(x), (int)Math.round(y));  }


  public static Point2D.Double midPt(Point2D.Double pt1, 
                                     Point2D.Double pt2)
  {  return new Point2D.Double((pt1.x + pt2.x)/2, (pt1.y + pt2.y)/2);  }


  // ------------ test rig -----------

  public static void main(String args[])
  {
    double[] vals = {0, 50, 90, 50, 90, 50 };
    ArrayList<Point2D.Double> pts = ShapeOps.tilePoints(10, 10, vals);

    System.out.println("Pts: " + Arrays.toString(pts.toArray()));
    
  }  // end of main()


}  // end of ShapeOps class
