
// Tile.java
// Andrew Davison, May 2022, ad@fivedots.coe.psu.ac.th

/* Immutable Tile data structure
   
   A tile consists of:
     * an image;
     * points represents the 'important coordinates' of the image
     * inner points (inside the image)
     * outer points (outside the image)

  A tile can be initialized using a Java shape, a Tile text file, an image, or
  an SVG file.

  All tile points are maintained as polar coordinates relative to the center 
  of the image.

  When using inner and outer points, it is assumed that the image points
  are specified in ccw order so that 'inside' the shape is to the left as
  the image points are traversed.

  ----
  Inner points are used for collision detection between tiles.
  Outer points are used to detect if another tile is adjacent to this tile.

  The detection algorithms rely on the underlying screen image being transparent, and
  on the background of the tile image also being transparent. 
  When the opaque tile is drawn on the screen image, that
  drawing area is no longer transparent, which is easy to detect.

  Inner and outer points can be supplied by the user, or will be generated
  automatically when the detection code is called (hasOverlap() for inners 
  and checkSides() for outers).

  The generated inners and outers are created by converting the image points
  into a shape and scaling it. This approach may break if the image/shape
  is concave.

  ---
  The main tile transformations are:
    * Rotation (around the center of the image).
      The angle is specified in degrees, running ccw from the +x-axis;

    * Resizing;

    * flipping horizontally or vertically around the center of the image.
      Note: flipping a tile will break the ccw ordering of the image points, 
      and so should not be used with inner and outer points;

    * changing the color on the tile, based on changing all the pixels
      that match the old color to the new color;

  Drawing
  -------
  By default the center of the tile's image is drawn at a specified (x,y) 
  position in a 'screen' image.
  
  Or you can use drawAt() to specify that a the specified image
  point is drawn at (x,y) in the 'screen' image.

  tryDrawAt() draws the tile to a new image so that the result can be examined
  without affecting the 'screen' image. testDrawAt() succeeds or fails. If it
  succeeds then it updates the screen image automatically.

  A tile can be viewed in a sperate JFrame, which is managed by the TileViewer class.
  There is also textual output with reportPoints(), or the simpler toString().

  Tiling
  ------
  The draw operations can be used to tile row-by-row, or there are two builtin
  search functions -- tileLocs() and tileSpacey() (defined in SEarch.java). 
  These builtins utilize inner and outer points, *and* point pairs.

  A point pair specifies a image point that can be paired with another point
  to allow two shapes to be placed side-by-side. A pair consists of a 'from'
  point paired with a 'to' point. A 'from' point specifies a side in ccw order.
  A 'to' point specifies a side in clockwise order.
*/

import javax.swing.*;
import java.awt.*;
import java.awt.geom.*;
import java.awt.event.*;
import java.awt.image.*;
import java.util.*;
import java.io.*;
import java.util.concurrent.*;


public class Tile
{
  // size of a tile
  public static final int WIDTH = 100;     
  public static final int HEIGHT = 100;

  public static final int MAX_SVG_POINTS = 30;

  private static final int TILE_BOUNDARY = 10;   // for SVG path output

  private static final int X_OFFSET = 100;   // for positioning windows in view()
  private static final int Y_OFFSET = 250; 

  private static final double INNER_SCALE = 0.6; 
  private static final double OUTER_SCALE = 1.4; 

  public static final double LENGTH_EPS = 3.0;  // in pixels

  private static String TILE_NM = "tile$$";
  private static String OPT_NM = "opt$$";   // outer point name prefix
  private static int idNum = 1;   // used to uniquely name cloned tiles and outer pts


  // window positioning in view() will not always be at the same coordinate
  private static int xWin, yWin;   // drawing coordinate
  private static int scrWidth, scrHeight;

  static {
    GraphicsEnvironment ge = GraphicsEnvironment.getLocalGraphicsEnvironment();
    GraphicsDevice gd = ge.getDefaultScreenDevice();
    Rectangle rect = gd.getDefaultConfiguration().getBounds();
    scrWidth = (int)Math.round(rect.getWidth());
    scrHeight = (int)Math.round(rect.getHeight());
    xWin = scrWidth - 200;  // initial drawing coordinate
    yWin = 10;
  }


  private String id;
  private boolean hasUserID = false;
  private BufferedImage im;
  private Shape shape = null;

  private HashMap<String,Polar> polars = new HashMap<>();
          // coordinates of the tile relative to the center of the image

  private ArrayList<String> ptNames = new ArrayList<>();
      /* the order is per the calls to addPt();
         Note: the order is NOT reversed after a flip
      */

  private boolean hasInners = false;
  private double innerScale = INNER_SCALE; 
  private ArrayList<Polar> innerPolars = new ArrayList<>();
           // used for detecting overlapping with other tiles

  private boolean hasOuters = false;
  private double outerScale = OUTER_SCALE; 
  private HashMap<String,Polar> outerPolars = new HashMap<>();
  private HashMap<String, Pair<String,String>> opSides = new HashMap<>();
           // used for detecting 'adjacent' tiles which mean
           // that the associated side is used
  private ArrayList<Pair<String,String>> usedSides = new ArrayList<>();


  // the center of image on the DRAWING SURFACE (screen/pane)
  private double xCenter, yCenter;
  private boolean isDrawn = false;



  public Tile(String fnm)
  {
    String nm = Pics.getName(fnm);
    this.id = (nm == null) ? TILE_NM+idNum++ : nm;
    if (nm != null)
      hasUserID = true;
    loadFile(fnm);   // id used for naming points
  }

  public Tile(String fnm, String id)
  {
    this.id = id;
    hasUserID = true;
    loadFile(fnm);
  }

  public Tile(BufferedImage im)
  { this.im = im;
    this.id = TILE_NM+idNum++;
  }

  public Tile(BufferedImage im, String id)
  { this.im = im;
    this.id = id;
    hasUserID = true;
  }


  public Tile(Shape s)
  {
    this.id = TILE_NM+idNum++;
    shapeToTile(s);
  }

  public Tile(Shape s, String id)
  {
    this.id = id;
    hasUserID = true;
    shapeToTile(s);   // id used for naming points
  }


  private void loadFile(String fnm)
  {
    String ext = Pics.getExtension(fnm);
    if (ext.equals("txt")) {
      ArrayList<Shape> shapes = ShapeOps.readShapes(fnm);
      shapesToTile(shapes);
    }
    else if (ext.equals("svg")) {
      ArrayList<Shape> shapes = ShapeOps.loadSVGPaths(fnm);
      shapesToTile(shapes);
      if (numPts() > MAX_SVG_POINTS)
        clearPts();   // removes excessive SVG path points
    }
    else if (ext.equals("path")) {
      Shape s = ShapeOps.loadPath(fnm);
      shapeToTile(s);
    }
    else
      im = Pics.load(fnm);
  }  // end of loadFile()



  private void shapesToTile(ArrayList<Shape> shapes)
  {
    Shape s = null;
    if (shapes.size() == 0) {
      System.out.println("No path shape found");
      return;
    }

    if (shapes.size() > 1) {
      System.out.println("Found " + shapes.size() + 
                          " path shapes; merging them into one");
      s = ShapeOps.combine(shapes);
    }
    else
      s = shapes.get(0);
    shapeToTile(s);
  }  // end of shapesToTile()



  private void shapeToTile(Shape s)
  {
    Path2D.Double path = new Path2D.Double();
    path.append(s, false);

    this.shape = path;

    Rectangle rect = path.getBounds();
    // System.out.println("bounded box: " + rect);
    
    int scrWidth = (rect.x + rect.width) + 2*TILE_BOUNDARY;
    int scrHeight = (rect.y + rect.height) + 2*TILE_BOUNDARY;
    BufferedImage shapeIm = new BufferedImage(scrWidth, scrHeight, 
                              BufferedImage.TYPE_INT_ARGB); // alpha channel
    Graphics2D g2d = Pics.createGraphics(shapeIm);
    ShapeOps.draw(g2d, path);
    g2d.dispose();
    this.im = shapeIm;

    ArrayList<Point2D.Double> pts = ShapeOps.getPoints(path);

    if (hasUserID)
      addPts(""+id.toUpperCase().charAt(0), pts);  // so label is short
    else
      addPts("S", pts);
  }  // end of shapeToTile()




  // ---------- corner point ops ------------------------


  public void addPt(String name, Point2D.Double pt)
  {  addPt(name, pt.x, pt.y);  }


  public void addPt(String name, double xIm, double yIm)
  // specify a point in terms of image pixels (i.e. top-left == (0,0))
  {
    // cartesian coord relative to center of image
    double x = xIm - im.getWidth()/2;
    double y = im.getHeight()/2 - yIm;

    // convert and save as Polar
    addPt(name, ShapeOps.cart2Polar(x, y));
  }  // end of addPt()



  public void addPt(String name, Polar p)
  {
    if (polars.containsKey(name))
      System.out.println("Updating " + name + " pt");
    else {
      //  System.out.println("Adding " + name + " pt");
      polars.put(name, p);
      ptNames.add(name);
    }
  }


  public void addPts(String name, ArrayList<Point2D.Double> pts)
  {  addPts(name, pts.toArray(new Point2D.Double[0]));  }


  public void addPts(String name, Point2D.Double[] pts)
  {  
    for (int i=0; i < pts.length; i++)
      addPt(name+i, pts[i].x, pts[i].y);
  }  // end of addPts()


  public void clearPts()
  {  
    System.out.println("Removing " + ptNames.size() + " points");
    polars.clear();
    ptNames.clear();
  }



  public ArrayList<Point2D.Double> getPtLocs()
  // get all the pt locations in the tile w.r.t the tile image coords
  {
    if (!isDrawn) {
      System.out.println("getPtLocs(): Tile " + id + 
                         " has not been drawn");
      return null;
    }

    ArrayList<Point2D.Double> pts = new ArrayList<>();
    for(String ptName : ptNames) {
      Point2D.Double pt = polars.get(ptName).getCartesian();
      double x = xCenter + pt.x;
      double y = yCenter - pt.y;
      pts.add( new Point2D.Double(x,y));
    }
    return pts;
  }  // end of getPtLocs()



  public Point2D.Double getPtLoc(String name)
  // get location of the pt on the screen
  {
    if (!polars.containsKey(name)) {
      System.out.println("No \"" + name + "\" pt for tile " + id);
      return null;
    }
    else {
      Point2D.Double pt = polars.get(name).getCartesian();
      if (!isDrawn) {
        System.out.println("getPtLoc(): Tile " + id + 
                           " has not been drawn");
        return null;
      }
      else {
        double x = xCenter + pt.x;
        double y = yCenter - pt.y;
        return new Point2D.Double(x,y);
      }
    }
  }  // end of getPtLoc()



  public ArrayList<Point2D.Double> getImPts()
  // get all the pts in the tile w.r.t the tile image coords
  {
    ArrayList<Point2D.Double> pts = new ArrayList<>();
    for(String ptName : ptNames) {
      Point2D.Double pt = polars.get(ptName).getCartesian();
      double x = im.getWidth()/2 + pt.x;
      double y = im.getHeight()/2 - pt.y;
      pts.add( new Point2D.Double(x,y));
    }
    return pts;
  }  // end of getImPts()


  public Point2D.Double getPtIm(String ptName)
  // get the pt in the tile w.r.t the tile image coords
  {
    if (!polars.containsKey(ptName)) {
      System.out.println("No \"" + ptName + "\" pt for tile " + id);
      return null;
    }
    else {
      Point2D.Double pt = polars.get(ptName).getCartesian();
      double x = im.getWidth()/2 + pt.x;
      double y = im.getHeight()/2 - pt.y;
      return new Point2D.Double(x,y);
    }
  }  // end of getPtIm()



  public Point2D.Double getCenterIm()
  // get the center pt for all the image pts
  {  
    double xc = 0;
    double yc = 0;
    for(String ptName : ptNames) {
      Point2D.Double pt = polars.get(ptName).getCartesian();
      xc += pt.x;
      yc += pt.y;
    }
    int n = ptNames.size();
    double x = im.getWidth()/2 + xc/n;
    double y = im.getHeight()/2 - yc/n;
    return new Point2D.Double(x,y);
  } // end of getCenterIm()



  public Point2D.Double getImCenter()
  // get the center pt of the tile image
  {  return new Point2D.Double(im.getWidth()/2,im.getHeight()/2);  } 



   public double angleTo(String p1Name, String p2Name)
   // angle from p1 to p2 points in degrees
   {
     if (!polars.containsKey(p1Name)) {
       System.out.println("No \"" + p1Name + "\" pt for tile " + id);
       return 0;
     }
     if (!polars.containsKey(p2Name)) {
       System.out.println("No \"" + p2Name + "\" pt for tile " + id);
       return 0;
     }

     Point2D.Double p1 = polars.get(p1Name).getCartesian();
     Point2D.Double p2 = polars.get(p2Name).getCartesian();
     return ShapeOps.getAngle(p1, p2);
   }  // end of angleTo()


   public double angle(String ptName)
   // angle in degrees between prevPt, ptName, and nextPt in ccw order
   {
     if (!polars.containsKey(ptName)) {
       System.out.println("No \"" + ptName + "\" pt for tile " + id);
       return 0;
     }

     Point2D.Double prevPt = polars.get(nextName(ptName)).getCartesian();
     Point2D.Double pt = polars.get(ptName).getCartesian();
     Point2D.Double nextPt = polars.get(prevName(ptName)).getCartesian();

     return ShapeOps.pointAngle(prevPt, pt, nextPt);
   }  // end of angleTo()



   public double length(Pair<String,String> side)
   {  return length(side.getX(), side.getY());  }


   public double length(String p1Name, String p2Name)
   // distance between p1 and p2
   {
     if (!polars.containsKey(p1Name)) {
       System.out.println("No \"" + p1Name + "\" pt for tile " + id);
       return 0;
     }
     if (!polars.containsKey(p2Name)) {
       System.out.println("No \"" + p2Name + "\" pt for tile " + id);
       return 0;
     }

     Point2D.Double p1 = polars.get(p1Name).getCartesian();
     Point2D.Double p2 = polars.get(p2Name).getCartesian();
     return ShapeOps.getDistance(p1, p2);
   }  // end of length()



  public void reportPoints()
  {
    System.out.println("Points for tile " + id + ":");
    if (numPts() == 0)
      System.out.println("  None defined");
    else {
      for (int i=0; i < numPts(); i++) {
        String ptName = ptName(i);
        String nextName = nextName(ptName);
        Point2D.Double pt = getPtIm(ptName);
        System.out.printf("  %s (%.2f, %.2f); %.2f degs; %.2f pixels\n", 
                   ptName, pt.x, pt.y,
                   angle(ptName), length(ptName, nextName));
      }
    }
    System.out.println("Image size: (" + getWidth() + ", " +
                                         getHeight() + ")");
    System.out.println("No. Inners: " + innerPolars.size() + 
                     "; No. Outers: " + outerPolars.size());
    System.out.println("---");
  }  // end of reportPoints()



  public Polar getPolar(String name)
  // get polar location of the pt
  {
    if (!polars.containsKey(name)) {
      System.out.println("No \"" + name + "\" pt for tile " + id);
      return null;
    }
    else
      return polars.get(name);
  }  // end of getPolar()


  public Point2D.Double getCartesian(String name)
  // get cartesian location of the pt
  {
    if (!polars.containsKey(name)) {
      System.out.println("No \"" + name + "\" pt for tile " + id);
      return null;
    }
    else
      return polars.get(name).getCartesian();
  }  // end of getCartesian()


 // public HashMap<String,Polar> getPolars()
 // {  return polars;  }


  public ArrayList<String> getPtNames()
  {  return ptNames;  }


  public int numPts()
  {  return ptNames.size();  }


  public String ptName(int i)
  {
    if ((i < 0) || (i > ptNames.size()-1)) {
      System.out.println("Point2D.Double index (" + i + ") out of range in tile " + id);
      return null;
    }
    else
      return ptNames.get(i);
  }


  public String nextName(String nm)
  {
    int pos = findNamePos(nm);
    if (pos == -1) {
      System.out.println("No point named " + nm + " in tile " + id);
      return null;
    }
    int nextPos =  (pos+1)%ptNames.size();
    // System.out.println("nextPos: " + nextPos);
    return ptNames.get(nextPos);
  }  // end of nextName()


  public String prevName(String nm)
  {
    int pos = findNamePos(nm);
    if (pos == -1) {
      System.out.println("No point named " + nm + " in tile " + id);
      return null;
    }
    int prevPos = Math.floorMod(pos-1, ptNames.size());
    // System.out.println("prevPos: " + prevPos);
    return ptNames.get(prevPos);
  }  // end of prevName()


  public int findNamePos(String nm)
  {
    for (int i=0; i < ptNames.size(); i++)
      if (ptNames.get(i).equals(nm))
        return i;
    return -1;
  }


  public boolean containsLoc(Point2D.Double pt)
  // does the drawn tile contain this screen point?
  {
    if (!isDrawn) {
      System.out.println("containsLoc(): Tile " + id + 
                         " has not been drawn");
      return false;
    }

    Shape tileShape = ShapeOps.points( getPtLocs());
    return tileShape.contains(pt);
  }  // end of contains()



  public Pair<String,String> nearestSide(Point2D.Double ptLoc)
  // which is side is ptLoc nearest to in the drawn tile?
  {
    if (!isDrawn) {
      System.out.println("nearestSide(): Tile " + id + 
                         " has not been drawn");
      return null;
    }

    ArrayList<Point2D.Double> ptLocs = getPtLocs();

    // use side from last pt to the first as starting distance
    Pair<String,String> minSide = new Pair<>(
                                 ptNames.get(ptLocs.size()-1),
                                 ptNames.get(0));
    Point2D.Double midPt = ShapeOps.midPt(
                                ptLocs.get(ptLocs.size()-1), 
                                ptLocs.get(0));
    double minDist = ShapeOps.getDistance(ptLoc, midPt);
    // System.out.println("minDist: " + minDist);
    
    for (int i=0; i < ptLocs.size()-1; i++) {
      Pair<String,String> side = new Pair<>(ptNames.get(i), 
                                            ptNames.get(i+1));
      midPt = ShapeOps.midPt(ptLocs.get(i), ptLocs.get(i+1));
      double dist = ShapeOps.getDistance(ptLoc, midPt);
      // System.out.println("dist: " + dist);
      if (dist < minDist) {
        minDist = dist;
        // System.out.println("minDist now: " + minDist);
        minSide = side;
      }
    }
    return minSide;
  }  // end of nearestSide()



  
  // --------------- inner points ----------------


  public void setInnerScale(double v)
  {
    if ((v <= 0) || (v >= 1)) {
      System.out.println("Inner scale must be between 0 and 1");
      return;
    }
    innerScale = v; 
  }

  public double getInnerScale()
  {  return innerScale;  }


  public void addInner(double xIm, double yIm)
  {
    // System.out.println("Add inner: (" + xIm + ", " + yIm + ")");

    // cartesian coord relative to center of image
    double x = xIm - im.getWidth()/2;
    double y = im.getHeight()/2 - yIm;

    // convert to Polar
    innerPolars.add( ShapeOps.cart2Polar(x, y));
    hasInners = true;
  }  // end of addInner()


  public void addInners(double[] xs, double[] ys)
  {
    int numPts = (xs.length < ys.length) ? xs.length : ys.length;
    if (xs.length != ys.length)
      System.out.println("Unequal array lengths; using smaller");

    for (int i=0; i < numPts; i++)
      addInner(xs[i], ys[i]);
  }  // end of addInners()


  public void addInners(ArrayList<Point2D.Double> inPts)
  {
    for (int i=0; i < inPts.size(); i++)
      addInner(inPts.get(i).x, inPts.get(i).y);
  } 


  public void addPolarInners(ArrayList<Polar> inPs)
  {
    for (Polar p : inPs)
      addInner(p);
  } 

  public void addInner(Polar p)
  {  innerPolars.add(p); 
     hasInners = true;
  }


  public boolean hasOverlap(BufferedImage im)
  {
    if (!hasInners) {
      // System.out.println("hasOverlap(): generating inner points for tile " + id);
      genInners();
    }

    for (Polar p : innerPolars) {
      if (hasOverlap(p, im))
        return true;
    }
    return false;
  }  // end of hasOverlap()


  private boolean hasOverlap(Polar p, BufferedImage im)
  {
    if (!isDrawn) {
     // System.out.println("hasOverlap(): Tile " + id + 
     //                    " has not been drawn");
      return false;
    }

    Point2D.Double localPt = p.getCartesian();
    int xScr = (int)Math.round(xCenter + localPt.x);
    int yScr = (int)Math.round(yCenter - localPt.y);
    boolean isOpaque = !Pics.isTransparent(im, xScr, yScr);
/*
    if (isOpaque) {
      System.out.println("pt: " + localPt + 
               " at (" + xScr + ", " + yScr + ")");
      Graphics2D g2d = Pics.createGraphics(im);
      g2d.setColor(Color.BLUE);
      g2d.fillOval(xScr-3, yScr-3, 6, 6); 
    }
*/
    return isOpaque;
  }  // end of hasOverlap()


  public void genInners()
  /* Generate inner pts for the tile by creating a shape from
     its points, scaling it down. Convert the smaller shape into a list
     of points, with some extra points added along each side.

     This approach will fail if the original shape is concave.
  */
  {
    if (hasInners) {
      System.out.println("Tile already has inner points");
      return;
    }

    ArrayList<Point2D.Double> shapePts = getImPts();
    if ((shapePts == null) || shapePts.isEmpty())
      return;
    // System.out.println("shape pts: " + Arrays.toString( shapePts.toArray()));

    Shape shape = ShapeOps.points(shapePts);
    Shape smallShape = ShapeOps.resize(shape, innerScale);
    ArrayList<Point2D.Double> iPts = ShapeOps.genExtras(smallShape, 1);
                                   // no. of points added per side
    discardExternals(iPts, shape);
  
    addInners(iPts);
    hasInners = true;
  }  // end of genInners()


  private void discardExternals(ArrayList<Point2D.Double> iPts,
                                Shape s)
  // any points which are not inside the shape/image are deleted
  {
    int numDeleted = 0;

    // check if any points not inside shape
    int idx =0;	
    while (idx < iPts.size()) {
      if (!s.contains( iPts.get(idx))) {
        iPts.remove(idx);
        numDeleted++;
      }
      else
        idx++;
    }

    // check if any points not inside image
    idx =0;
    while (idx < iPts.size()) {
      Point2D.Double ipt = iPts.get(idx);
      if (Pics.isTransparent(im, ipt.x, ipt.y)) {
        // System.out.println("pt (" + ipt.x + ", " + ipt.y + ")");
        iPts.remove(idx);
        numDeleted++;
      }
      else
        idx++;
    }

    if (numDeleted > 0) {
      System.out.println("WARNING: deleted " + numDeleted + 
                         " poorly positioned inner points");
      System.out.println("  No. of remaining inners: " + iPts.size() + "\n");
    }
  }  // end of discardExternals()



  public ArrayList<Point2D.Double> getInnerImPts()
  // get locations of all the inner pts on the image
  {
    ArrayList<Point2D.Double> pts = new ArrayList<>();
    for (Polar p : innerPolars) {
      Point2D.Double pt = p.getCartesian();
      double x = im.getWidth()/2 + pt.x;
      double y = im.getHeight()/2 - pt.y;
      pts.add( new Point2D.Double(x,y) );
    }
    return pts;
  }  

  
  // --------------- outer points ----------------


  public void setOuterScale(double v)
  {
    if ((v <= 1) || (v >= 2)) {
      System.out.println("Outer scale must be between 1 and 2");
      return;
    }
    outerScale = v; 
  }

  public double getOuterScale()
  {  return outerScale;  }


  public void addOuter(String ptNm1, String ptNm2, Point2D.Double pt)
  {  addOuter(ptNm1, ptNm2, pt.x, pt.y);  }


  public void addOuter(String ptNm1, String ptNm2, double xIm, double yIm)
  // add maps of outerpoint to a side (ptNm1--ptNm2) and a Polar coordinate
  {
    // System.out.println("Add outer: (" + xIm + ", " + yIm + ")");

    // cartesian coord relative to center of image
    double x = xIm - im.getWidth()/2;
    double y = im.getHeight()/2 - yIm;

    int pos1 = findNamePos(ptNm1);
    if (pos1 == -1) {
      System.out.println("No point named " + ptNm1 + " in tile " + id);
      return;
    }

    int pos2 = findNamePos(ptNm2);
    if (pos2 == -1) {
      System.out.println("No point named " + ptNm2 + " in tile " + id);
      return;
    }

    String optNm = OPT_NM + idNum;
    idNum++;
    opSides.put(optNm, new Pair<>(ptNm1, ptNm2));
    outerPolars.put(optNm, ShapeOps.cart2Polar(x, y));

    hasOuters = true;
  }  // end of addOuter()


  public void addOuter(String optNm, Pair<String,String> side, Polar p)
  {
    opSides.put(optNm, side);
    outerPolars.put(optNm, p);
    hasOuters = true;
  }



  public void genOuters()
  /* Generate outer pts for the tile by creating a shape from
     its points, and scaling it up. The midpoint of each side
     of this bigger shape are used as the outer points.

     This approach may fail if the original shape is concave.
  */
  {
    if (hasOuters) {
      System.out.println("Tile already has outer points");
      return;
    }

    ArrayList<Point2D.Double> shapePts = getImPts();
    if ((shapePts == null) || shapePts.isEmpty())
      return;

    Shape shape = ShapeOps.points(shapePts);
    Shape bigShape = ShapeOps.resize(shape, outerScale);

    // discard points which are inside shape or image
    int numDiscarded = 0;
    for (int i=0; i < numPts(); i++) {
      Point2D.Double midPt = ShapeOps.getMidPoint(bigShape, i);

      if (shape.contains(midPt))   // pt in shape
        numDiscarded++;
      else if (!Pics.isTransparent(im, midPt.x, midPt.y)) {  // pt inside image
        // System.out.println("pt (" + midPt.x + ", " + midPt.y + ")");
        numDiscarded++;
      }
      else {
        String ptNm = ptName(i);
        String nextNm = nextName(ptNm);
        addOuter(ptNm, nextNm, midPt);
      }
    }

    if (numDiscarded > 0) {
      System.out.println("WARNING: discarded " + numDiscarded + 
                         " poorly positioned outer points");
      System.out.println("  No. of remaining outers: " + 
                                   (numPts() - numDiscarded) + "\n");
    }

    hasOuters = true;
  }  // end of genOuters()



  public Point2D.Double findOuterLoc(String ptNm1, String ptNm2)
  // get location of the outerpoint pt on the screen
  {
    for (String optName : opSides.keySet()) {
      Pair<String,String> side = opSides.get(optName);
      if (isSideEqual(ptNm1, ptNm2, side)) {
        Point2D.Double pt = outerPolars.get(optName).getCartesian();
        double x = xCenter + pt.x;
        double y = yCenter - pt.y;
        return new Point2D.Double(x,y);
      }
    }
    return null;
  }  // end of findOuterLoc()



  public ArrayList<Point2D.Double> getOuterImPts()
  // get locations of all the outer pts on the image
  {
    ArrayList<Point2D.Double> pts = new ArrayList<>();
    for (String optName : outerPolars.keySet()) {
      Point2D.Double pt = outerPolars.get(optName).getCartesian();
      double x = im.getWidth()/2 + pt.x;
      double y = im.getHeight()/2 - pt.y;
      pts.add( new Point2D.Double(x,y) );
    }
    return pts;
  }  



  // -------------- used sides  -----------------


  public void checkSides(BufferedImage im)
  /* Use the tile's outer points to decide which sides
     are 'used' (i.e. is adjacent to another tile's side).
  */
  {
    if (!hasOuters) {
      // System.out.println("checkSides(): generating outer points for tile " + id);
      genOuters();
    }

    usedSides.clear();
    for (String optName : outerPolars.keySet()) {
      Polar p = outerPolars.get(optName);
      if (hasOverlap(p, im)) {
        Pair<String,String> side = opSides.get(optName);
        // System.out.println("CheckSide " + side + " used in tile " + id);
        usedSides.add(side);
      }
    }
  }  // end of checkSides()


  public void addSide(String ptNm1, String ptNm2)
  {
    int pos1 = findNamePos(ptNm1);
    if (pos1 == -1) {
      System.out.println("No point named " + ptNm1 + " in tile " + id);
      return;
    }

    int pos2 = findNamePos(ptNm2);
    if (pos2 == -1) {
      System.out.println("No point named " + ptNm2 + " in tile " + id);
      return;
    }

    if (Math.abs(pos1 - pos2) != 1) {
      System.out.println(ptNm1 + " and " + ptNm2 + " are not adjacent");
      return;
    }

    if (isUsedSide(ptNm1, ptNm2)) {
      System.out.println("Side " + ptNm1 + " -- " + ptNm2 + " used already");
      return;
    }

    usedSides.add( new Pair<String,String>(ptNm1, ptNm2));
  }  // end of addSide()



  public Pair<String,String> findFreeSide()
  /* return the (pt,nextPt) names of a free side when looking
     at the sides of the tile in ccw, or null
  */
  {
    for (String ptName : ptNames) {
       String nextName = nextName(ptName);
       if (!isUsedSide(ptName, nextName))
         return new Pair<String,String>(ptName, nextName);
    }
    return null;
  }


  public boolean isUsedSide(String ptNm1, String ptNm2)
  {  return (findSidePos(ptNm1, ptNm2) != -1);  }  


  public int findSidePos(String ptNm1, String ptNm2)
  {
    for(int i=0; i < usedSides.size(); i++) {
      Pair<String,String> side = usedSides.get(i);
      if (isSideEqual(ptNm1, ptNm2, side))
        return i;
    }
    return -1;
  }  // end of findSidePos()



  public boolean isSideEqual(String ptNm1, String ptNm2, 
                               Pair<String,String> side)
  { return ( (ptNm1.equals(side.getX()) && ptNm2.equals(side.getY())) ||
             (ptNm1.equals(side.getY()) && ptNm2.equals(side.getX())) );
  }



  public Line2D.Double side2Line(Pair<String,String> side)
  {  return side2Line(side.getX(), side.getY());  }


  public Line2D.Double side2Line(String ptNm1, String ptNm2)
  {
    Point2D.Double pt1 = getPtIm(ptNm1);
    if (pt1 == null)
      return null;

    Point2D.Double pt2 = getPtIm(ptNm2);
    if (pt2 == null)
      return null;

    return new Line2D.Double(pt1.x, pt1.y, pt2.x, pt2.y);
  }  // end of side2Line()



  public boolean checkSidesMatch(ArrayList<Tile> otherTiles,
                                 PointPairs ptPairs)
  {
    for (int i=0; i < numPts(); i++) {
      String ptName = ptName(i);
      String ptName1 = nextName(ptName);  // assume ccw order

      Point2D.Double oPt = findOuterLoc(ptName, ptName1);
      if (oPt != null) {
        Tile outerTile = Tile.nearest(oPt, otherTiles);
        if (outerTile != null) {
          
          // find nearest side of that tile to the point
          Pair<String,String> side = outerTile.nearestSide(oPt);
          // compare two side lengths
          if (Math.abs(outerTile.length(side) - 
                       length(ptName, ptName1)) > Tile.LENGTH_EPS)
            return false;
          // compare the match points pair
          if (!ptPairs.isMatch(ptName, side.getY()))
            return false;
        }
      }
    }  // end of for-loop
    return true;
  }  // end of checkSidesMatch()




  // -------------  other ops -------------------


  public void setID(String newID)
  {  id = newID;  }

  public String getID()
  {  return id;  }

  public Shape getShape()
  {  return shape;  }

  public BufferedImage getImage()
  {  return im;  }

  public Graphics2D getGraphics()
  {  return Pics.createGraphics(im);  }

  public int getWidth()
  {  return im.getWidth();  }

  public int getHeight()
  {  return im.getHeight();  }


  public boolean isDrawn()
  {  return isDrawn;  }

  public void unsetDrawn()
  {  isDrawn = false;  }


  public Point2D.Double getCenterLoc()
  // get location of center of image on the screen
  {
    if (!isDrawn) {
      System.out.println("getCenterLoc(): Tile " + id + 
                         " has not been drawn");
      return null;
    }
    else
      return new Point2D.Double(xCenter, yCenter);
  }


  public String toString()
  {
    if (!isDrawn)
      return "<tile: " + id + " " + polars.toString() + ">";
    else
      return "<tile: " + id + " at (" + xCenter + "," +
                  + yCenter + ") " + polars.toString() + ">";
  }


  // -------------- tile transormations --------------------------


  public Tile clone()
  {
    Tile clone = new Tile(im, id + "--" + idNum);
    idNum++;

    for (String ptName : ptNames) {
      Polar p = polars.get(ptName);
      clone.addPt(ptName, p);
    }

    clone.addPolarInners(innerPolars);

    for (String optName : outerPolars.keySet()) {
      Pair<String,String> side = opSides.get(optName);
      Polar p = outerPolars.get(optName);
      clone.addOuter(optName, side, p);
    }

    return clone;
  }  // end of clone()



  public Tile resize(double scale)
  {
    if (scale <= 0) {
      System.out.println("Tile " + id + ": scale factor must be positive");
      return this;
    }

    BufferedImage imNew = Pics.resize(im, scale);
    Tile clone = new Tile(imNew, id + "--" + idNum);
    idNum++;

    for (String ptName : ptNames) {
      Polar p = polars.get(ptName);
      clone.addPt(ptName, p.resize(scale));
    }

    for (Polar ip : innerPolars)
      clone.addInner(ip.resize(scale));

    for (String optName : outerPolars.keySet()) {
      Pair<String,String> side = opSides.get(optName);
      Polar p = outerPolars.get(optName);
      clone.addOuter(optName, side, p.resize(scale));
    }

    return clone;
  }  // end of resize()



  public Tile rotate(double angle)
  // rotate the tile ccw by angle degrees
  {
    if ((angle < 0) || (angle > 360))
      angle = restrictAngle(angle);

    BufferedImage imNew = Pics.rotate(im, angle);
    Tile clone = new Tile(imNew, id + "--" + idNum);
    idNum++;

    for (String ptName : ptNames) {
      Polar p = polars.get(ptName);
      clone.addPt(ptName, p.rotate(angle));   // ccw
    }

    for (Polar ip : innerPolars)
      clone.addInner(ip.rotate(angle));

    for (String optName : outerPolars.keySet()) {
      Pair<String,String> side = opSides.get(optName);
      Polar p = outerPolars.get(optName);
      clone.addOuter(optName, side, p.rotate(angle));
    }

    return clone;
  }  // end of rotate()


  private double restrictAngle(double d)
  // d must range between 0 and 360
  {
    if (d > 360) {
      while (d > 360)
        d =  d - 360;
    }

    if (d < 0) {
      while (d < 0)
        d = d + 360;
    }
    return d;
  }  // end of restrictAngle()



  public double edgeRotation(String nPtName,
                        Tile drawnTile, String drawnPtName)
  /* calculate necessary rotation of this tile
     so that nPtName can be drawn next to drawnPtName
     and the edge nPtName --> nPtName-1 lines up 
     with the edge drawnPtName --> drawPtName+1.

     For drawnTile, the two points must be in ccw order (++)
     For new tile, the two points must be in cw order (--)
  */
  {
    String next = drawnTile.nextName(drawnPtName);
    double dAngle = drawnTile.angleTo(drawnPtName, next);
             // angle from dPt to dPt1
    // System.out.println("dAngle: " + dAngle);

    String prev = prevName(nPtName);
    double nAngle = angleTo(nPtName, prev);
             // angle from nPt to nPt-1 
    // System.out.println("nAngle: " + nAngle);

    double angle = dAngle - nAngle;
    if (angle < 0)
      angle += 360;
   // System.out.println("Rotate " + nPtName + " to " + 
   //                       drawnPtName + ": " +  angle);
    return angle;
  }  // end of edgeRotation()



  public Tile flip(int flipType)
  {
    if ((flipType == Pics.VERT) || (flipType == Pics.HORIZ) ||
        (flipType == Pics.BOTH)) {
      BufferedImage imNew = Pics.flip(im, flipType);
      Tile clone = new Tile(imNew, id + "--" + idNum);
      idNum++;

   //   ArrayList<String> revNames = new ArrayList<>(ptNames);
   //   Collections.reverse(revNames);

      for (String ptName : ptNames) {
        Polar p = polars.get(ptName);
        clone.addPt(ptName, p.flip(flipType));
      }

      for (Polar ip : innerPolars)
        clone.addInner(ip.flip(flipType));

      for (String optName : outerPolars.keySet()) {
        Pair<String,String> side = opSides.get(optName);
        Polar p = outerPolars.get(optName);
        clone.addOuter(optName, side, p.flip(flipType));
      }

      return clone;
    }
    else {
      System.out.println("Tile " + id + ": unknown flip type: " + flipType);
      return this;
    }
  }  // end of flip()



  public Tile colorChg(Color newCol)
  {  return colorChg(Color.WHITE, newCol, Pics.MAX_COLOR_DIST);  }


  public Tile colorChg(Color currCol, Color newCol)
  {  return colorChg(currCol, newCol, Pics.MAX_COLOR_DIST);  }


  public Tile colorChg(Color currCol, Color newCol, double maxDist)
  {
    BufferedImage imNew = Pics.colorChg(im, currCol, newCol, maxDist);
    Tile clone = new Tile(imNew, id + "--" + idNum);
    idNum++;

    for (String ptName : ptNames) {
      Polar p = polars.get(ptName);
      clone.addPt(ptName, p);
    }

    clone.addPolarInners(innerPolars);

    for (String optName : outerPolars.keySet()) {
      Pair<String,String> side = opSides.get(optName);
      Polar p = outerPolars.get(optName);
      clone.addOuter(optName, side, p);
    }

    return clone;
  }  // end of colorChg()


  // ------------------- drawing ops ----------------


  public void drawAt(BufferedImage im, String ptName)
  {  drawAt(im, ptName, im.getWidth()/2, im.getHeight()/2);  }


  public void drawAt(BufferedImage im, String ptName, Point2D.Double p)
  {  drawAt(im, ptName, p.x, p.y);  }


  public void drawAt(BufferedImage im, String ptName, 
                     double x, double y)
  {
    Graphics2D g2d = Pics.createGraphics(im);
    drawAt(g2d, ptName, x, y);
  }


  private void drawAt(Graphics2D g2d, String ptName, 
                      double x, double y)
  // draw the image so that pt "ptName" is at (x,y) in the pane
  {
    Point2D.Double imPt = getPtIm(ptName);
    if (imPt == null)
      return;

    // draw top-left of image pt offset from (x,y)
    g2d.drawImage(im, (int)Math.round(x-imPt.x), 
                      (int)Math.round(y-imPt.y), null);

    // record image center's position on the pane
    xCenter = x - imPt.x + im.getWidth()/2;
    yCenter = y - imPt.y + im.getHeight()/2;

   // if (isDrawn)
   //   System.out.println("Tile " + id + " being redrawn");
    isDrawn = true;
  }  // end of drawAt()



  public void draw(BufferedImage im)
  {  draw(im, im.getWidth()/2, im.getHeight()/2);   }

  public void draw(BufferedImage im, Point2D.Double pt)
  {  draw(im, pt.x, pt.y);   }

  public void draw(BufferedImage im, double x, double y)
  {
    Graphics2D g2d = Pics.createGraphics(im);
    draw(g2d, x, y);
  }

  private void draw(Graphics2D g2d, double x, double y)
  // draw the image so that the center of the image is at (x,y) in the pane
  {
    // draw top-left of image offset so its center 
    // is at (x,y) in pane
    g2d.drawImage(im, (int)Math.round(x-im.getWidth()/2), 
                      (int)Math.round(y-im.getHeight()/2), null);

    // record image center's position on the pane
    xCenter = x;
    yCenter = y;

   // if (isDrawn)
   //   System.out.println("Tile " + id + " being redrawn");
    isDrawn = true;
  }  // end of draw()




  public BufferedImage tryDrawAt(BufferedImage im, String ptName)
  /* draw to a new image so that the result can be examined without
     affecting the old image */
  {  return tryDrawAt(im, ptName, im.getWidth()/2, im.getHeight()/2);  }


  public BufferedImage tryDrawAt(BufferedImage im, 
                                    String ptName, Point2D.Double pt)
  {  return tryDrawAt(im, ptName, pt.x, pt.y);  }


  public BufferedImage tryDrawAt(BufferedImage im, 
                                String ptName, double x, double y)
  {
    BufferedImage copyIm = Pics.copy(im);
    Graphics2D g2d = Pics.createGraphics(copyIm);
    drawAt(g2d, ptName, x, y);

    if(hasOverlap(im)) {    // look at image without shape
      // System.out.println("Overlap detected");
      isDrawn = false;
      return null;
    }
    else
      return copyIm;
  }   // end of tryDrawAt()



  public boolean testDrawAt(BufferedImage im, String ptName)
  {  return testDrawAt(im, ptName, im.getWidth()/2, im.getHeight()/2);  }


  public boolean testDrawAt(BufferedImage im, String ptName, Point2D.Double pt)
  {  return testDrawAt(im, ptName, pt.x, pt.y);  }


  public boolean testDrawAt(BufferedImage im, String ptName, double x, double y)
  {  
    BufferedImage newIm = tryDrawAt(im, ptName, x, y);
    if (newIm != null) {
      Pics.assign(im,newIm);
      return true;
    }
    else
      return false;
  }


  // ----------------- viewing a tile -------------------------------
  // utilizes the TileViewer class


  public TileViewer view()
  {  return view(id, null, true);  }

  public TileViewer view(String label)
  {  return view(label, null, true);  }


  public TileViewer view(String label, String ptName)
  {   return view(label, ptName, true);  }


  public TileViewer view(String label, String ptName, boolean saveImage)
  // display tile centered in a pane at ptName
  {
    if (!hasInners)
      genInners();
    if (!hasOuters)
      genOuters();

    TileViewer pane = new TileViewer(this, label, ptName, saveImage);

    CountDownLatch drawnSignal = new CountDownLatch(1);
       // used to delay return from this method until the frame has been drawn

    JFrame f = new JFrame("View Tile " + id);
    f.addWindowListener( new WindowAdapter() {
      public void windowOpened(WindowEvent e)
      {  drawnSignal.countDown();  }
    });

    f.getContentPane().add(pane, BorderLayout.CENTER);

    // f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    f.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);  // do not exit appl.

    f.pack();
    f.setResizable(false);

    // offset each new window from previous Tile windows
    f.setLocation(Tile.xWin - f.getWidth(), Tile.yWin);

    // change position for next view() call
    Tile.yWin += Tile.Y_OFFSET; 
    if (Tile.yWin > Tile.scrHeight - f.getHeight()) {
      Tile.yWin = 10;
      Tile.xWin -= Tile.X_OFFSET; 
      if (Tile.xWin < 0)
        Tile.xWin += Tile.scrWidth;
    }

    f.setVisible(true);

    // wait for jframe to be fully drawn
    try {
      drawnSignal.await();
    }
    catch(InterruptedException e) {}

    return pane;
  } // end of view()



  public void addLabel(String label)
  // useful for debugging
  {
    Point2D.Double cPt = getCenterIm();
    Graphics2D g2d = getGraphics();

    g2d.setColor(Color.BLUE);
    Font font = new Font("SanSerif", Font.BOLD, 16);
    g2d.setFont(font);

    g2d.drawString(label, (int)cPt.x, (int)cPt.y);
  }  // end of addlabel()

  

  // =============== static methods =================



  public static Tile nearest(Point2D.Double ptLoc, 
                          ArrayList<Tile> placedTiles)
  // which tile in the list is this point nearest to in the drawn image
  {
    for(Tile t : placedTiles) {
      if (t.containsLoc(ptLoc))
        return t;
    }
    return null;
  }





  // ------------ test rig -----------

  public static void main(String args[])
  {
    Tile t1 = new Tile("data/trif.png");
    t1.addPt("A", 51,6);
    t1.addPt("B", 4,86);
    t1.addPt("C", 97,86);
    t1.reportPoints();

    // display tile
    t1.view();
    System.out.println("\nCenter: " + t1.getCenterLoc());
    System.out.println("Point2D.Double B: " + t1.getPtLoc("B"));

    Tile t2 = t1.rotate(60);
    System.out.println("\nRotate 90 = " + t2);
    t2.view();

    t2 = t2.resize(2);
    System.out.println("\nResized by 2 = " + t2);

    t2 = t2.flip(Pics.HORIZ);
    System.out.println("\nFlipped horiz = " + t2);

    t2 = t2.flip(Pics.VERT);
    System.out.println("\nFlipped vert = " + t2);
    t2.view();

  } // end of main()

} // end of Tile