
// TileViewer.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, May 2022

/* Displays a tile's image along with labelled image points.
   Inner and outer points are also drawn as colored dots
   (green for inner, blue for outer).
   If the user hasn't supplied any of these points, then 
   default values are generated before TileViewer is called.

   The tile image is drawn in a larger window with extra border
   space so that the point labels can be seen clearly.

   The actual bounds of the image are indicated with a dashed 
   border.

   When the tile viewer is closed, a copy of the image (without
   point labels) is written to the file TileViewer.png.
   This can be loaded into a graphics program (e.g. MS Paint)
   so the pixel position of points can be more easily
   checked.

   It is also possible to click on the loaded image to see a popup
   window containing the cursor's current position inside the image.
*/


import java.awt.*;
import javax.swing.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.util.*;



public class TileViewer extends JPanel 
             implements MouseListener, MouseMotionListener

{ 
  private static final Color LIGHT_BLUE = new Color(173,216,230);
  private static final Color PINK = new Color(253,185,200);

  private static final int BORDER = 20;
     // pixel space around the displayed image

  private static final String VIEW_FNM = "TileViewer.png";
             // the tile image is always saved to this file at close time


  private int pWidth, pHeight;
  private Tile tile;
  private String label;
  private String ptName;
  private int xPos, yPos;  // top-left of image in pane

  private Popup coordsPopup;  // for showing cursor coordinates



  public TileViewer(Tile tile, String label, String ptName, boolean saveImage)
  {
    this.tile = tile;
    this.label = label;
    this.ptName = ptName;

    if (saveImage)
      Pics.save(tile.getImage(), VIEW_FNM);

    pWidth = tile.getWidth() + (2*BORDER);
    pHeight = tile.getHeight() + (2*BORDER);
    setPreferredSize( new Dimension(pWidth, pHeight));

    addMouseListener(this);
    addMouseMotionListener(this);
  }  // end of TileViewer()



  public void paintComponent(Graphics g)
  {
    super.paintComponent(g);
    Graphics2D g2d = (Graphics2D)g; 

    BufferedImage im = tile.getImage();

    // draw tile at center of pane
    if (ptName == null) {
      xPos = pWidth/2 - im.getWidth()/2;
      yPos = pHeight/2 - im.getHeight()/2;
      g2d.drawImage(im, xPos, yPos, null);
    }
    else  { // locate ptName at center 
      Point2D.Double imPt = tile.getPtIm(ptName);
      xPos = pWidth/2 - (int) Math.round(imPt.x);
      yPos = pHeight/2 - (int) Math.round(imPt.y);
      g2d.drawImage(im, xPos, yPos, null);
    }

    dashedBorder(g2d, xPos, yPos, im.getWidth(), im.getHeight());
       // shows the bounds of the image in the display pane

    Font font = new Font("SanSerif", Font.BOLD, 14);
    g2d.setFont(font);

    // the label of each image point is drawn inside a pink circle
    ArrayList<String> nms = tile.getPtNames();
    for(String nm : nms) {
      Point2D.Double pt = tile.getPtIm(nm);
      //int xPt = pWidth/2 - (int)Math.round(pt.x);
      //int yPt = pHeight/2 - (int)Math.round(pt.y);
      int xPt = xPos + (int)Math.round(pt.x);
      int yPt = yPos + (int)Math.round(pt.y);
      g2d.setColor(PINK);
      g2d.fillOval(xPt-10, yPt-10, 20, 20);
      g2d.setColor(Color.BLACK);
      Pics.centerText(g2d, nm, xPt, yPt);
    }

    // inner points are drawn as green dots
    ArrayList<Point2D.Double> iPts = tile.getInnerImPts();
    if (!iPts.isEmpty()) {
      for(Point2D.Double pt : iPts) {
        int xPt = xPos + (int)Math.round(pt.x);
        int yPt = yPos + (int)Math.round(pt.y);
        g2d.setColor(Color.GREEN);
        g2d.fillOval(xPt-3, yPt-3, 6, 6);
      }
    }

    // outer points are dran as blue dots
    ArrayList<Point2D.Double> oPts = tile.getOuterImPts();
    if (!oPts.isEmpty()) {
      for(Point2D.Double pt : oPts) {
        int xPt = xPos + (int)Math.round(pt.x);
        int yPt = yPos + (int)Math.round(pt.y);
        g2d.setColor(Color.BLUE);
        g2d.fillOval(xPt-3, yPt-3, 6, 6);
      }
    }

    // the name of the tile
    g2d.setColor(Color.BLACK);
    g2d.drawString(label, 5, pHeight-5);
  } // end of paintComponent()



  private void dashedBorder(Graphics2D g2, 
                       int x, int y, int w, int h)
  {
    Graphics2D g2d = (Graphics2D) g2.create();
   
    Stroke dashed = new BasicStroke(3, 
                BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL,
                 0, new float[]{5}, 0);
    g2d.setStroke(dashed);
    g2d.setPaint(Color.GRAY);
   
    g2d.drawRect(x, y, w, h);
    g2d.dispose();
  }



  // ------- MouseListener methods --------

  public void mousePressed( MouseEvent e)
  // display the mouse coordinates in a popup window
  { 
    JLabel text = new JLabel("(" + 
          (e.getX()-xPos) + "," + (e.getY()-yPos) + ")");
    coordsPopup = PopupFactory.getSharedInstance().
                  getPopup(e.getComponent(), text, 
                     e.getXOnScreen()+15, e.getYOnScreen());
    coordsPopup.show();
  }

  public void mouseReleased( MouseEvent e)
  { 
    if (coordsPopup != null)
      coordsPopup.hide();
  }

  public void mouseClicked(MouseEvent e) {}
  public void mouseEntered(MouseEvent e) {}
  public void mouseExited(MouseEvent e) {}


  // ------- MouseMotionListener methods --------

  public void mouseDragged( MouseEvent e)
  // display the mouse coordinates in a popup window
  {
   if (coordsPopup != null)
     coordsPopup.hide();
   JLabel text = new JLabel("(" + 
         (e.getX()-xPos) + "," + (e.getY()-yPos) + ")");
   coordsPopup = PopupFactory.getSharedInstance().
                 getPopup(e.getComponent(), text, 
                    e.getXOnScreen()+15, e.getYOnScreen());
   coordsPopup.show();
  }

  public void mouseMoved(MouseEvent e) {}

 
}  // end of TileViewer class