
// DrawFish.java
// Andrew Davison, May 2022, ad@fivedots.coe.psu.ac.th

/* Draw M.C. Escher's "Square Limit" (1964) down to three levels.
   https://www.wikiart.org/en/m-c-escher/square-limit
   https://www.nga.gov/collection/art-object-page.135604.html

   This code uses tile composition to build up to the final image.

   The component tiles are stored in a map (Name-->tile) for easy
   access.

*/

import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.util.*;


public class DrawFish
{
  private static final Color LIGHT_BLUE = new Color(173,216,230);
  private static final Color GRAY_GREEN = new Color(179, 204, 204);
  private static final Color LIGHT_BROWN = new Color(214, 214, 194);


  private static final double SCALE = Math.sqrt(2)/2;

  private static HashMap <String,Tile> tilesMap;


  public static void main(String args[])
  {
    createTilesMap();

    // the top-level tile (made from smaller tile pieces)
    Tile quarTile = makeQuarter();
    //quarTile.view();
    //quarTile.reportPoints();


    // build picture
    double horizDist = quarTile.length("C1", "C2");
    // System.out.println("\nHoriz dist in quarTile: " + horizDist);

    int pWidth = (int)Math.round( horizDist * 1.05);  
    int pHeight = (int)Math.round( horizDist * 1.05); 
           // adds a small border around the rendered image
    ImageViewer iview = Pics.view("Square Limit", pWidth, pHeight);
    BufferedImage im = iview.getImage(); 

    // color the background
    Graphics2D g2d = Pics.createGraphics(im);
    g2d.setColor(LIGHT_BLUE);
    g2d.fillRect(0, 0, pWidth, pHeight);

    // display 4 'quarters' around the center
    for (int i = 0; i < 4; i++) {
      Tile t = quarTile.rotate(90*i);
      t.drawAt(im, "C0", pWidth/2, pHeight/2);
      iview.repaint();
      Pics.pause(100);
    }
    
    Pics.save(im, "SavedPics/escherFish.png");
  } // end of main()



  // ----------- tile creation ------------------------


  private static void createTilesMap()
  {
    tilesMap = new HashMap<String,Tile>();

    Tile fish = new Tile("data/fish.png");
    fish.addPt("Head", 35,33);
    fish.addPt("Left", 35, 163);
    fish.addPt("Right", 100, 100);   // at the center of this image
    fish.addPt("Tail", 165, 163);
    fish.view();
    fish.reportPoints();

    // basic fish
    tilesMap.put("p0", fish);
    tilesMap.put("p90", fish.rotate(90).colorChg(Color.WHITE, GRAY_GREEN));

    Tile flippedFish = fish.flip(Pics.HORIZ);
    tilesMap.put("f45", flippedFish.rotate(45).colorChg(Color.WHITE, GRAY_GREEN));
    tilesMap.put("f135", flippedFish.rotate(135).colorChg(Color.WHITE, LIGHT_BROWN));
    tilesMap.put("f225", flippedFish.rotate(225));

    // composite tiles
    tilesMap.put("mid", makeMid());
    tilesMap.put("rightEnd", makeRightEnd());
    tilesMap.put("leftEnd", makeLeftEnd());

/*
    System.out.println("Tiles Map");
    for (String key : tilesMap.keySet()) {
       System.out.println("-- " + key + " ---");
       Tile t = tilesMap.get(key);
       t.view();
       t.reportPoints();
    }
*/
  }  // end of createTilesMap()



  private static Tile makeMid()
  {
    BufferedImage im = new BufferedImage(460, 260, 
                                 BufferedImage.TYPE_INT_ARGB);  // alpha channel

    Point2D.Double startPt = new Point2D.Double(15, 15);
       // near the top left point of the drawing area

    Point2D.Double[] corners = new Point2D.Double[4];


    // left side square of rectangle --------------

    // tile along the top of the square
    Tile top = tilesMap.get("f135").resize(SCALE);
    top.drawAt(im, "Tail", startPt);
    corners[3] = startPt;

    // tile up the left side of the square
    Tile left = tilesMap.get("f225").resize(SCALE);
    left.drawAt(im, "Head", startPt);
    corners[0] = left.getPtLoc("Tail");

    // tile in the bottom right corner of the square
    Tile right = tilesMap.get("p90"); 
    right.drawAt(im, "Head", corners[0]);

    Point2D.Double rightPt = right.getPtLoc("Tail");


    // right side square of rectangle -----------

    // tile along the top of the square
    top = tilesMap.get("f135").resize(SCALE);
    top.drawAt(im, "Tail", rightPt);
    corners[2] = top.getPtLoc("Head");

    // tile down the right side of the square
    right = tilesMap.get("f45").resize(SCALE);
    right.drawAt(im, "Tail", corners[2]);
    corners[1] = right.getPtLoc("Head");

    // tile in the bottom left corner of the square
    left = tilesMap.get("p0"); 
    left.drawAt(im, "Head", rightPt);

    Tile midTile = new Tile(im, "mid");
    for (int i=0; i < 4; i++)
      midTile.addPt("C"+i, corners[i]);

    return midTile;
  }   // end of makeMid()




  private static Tile makeRightEnd()
  // the right end triangle of the fishes tiles
  {
    BufferedImage im = new BufferedImage(300, 300, 
                                 BufferedImage.TYPE_INT_ARGB); 

    Point2D.Double startPt = new Point2D.Double(150, 150);
       // middle of the drawing area

    Point2D.Double[] corners = new Point2D.Double[3];

    // tile along the top of the triangle
    Tile top = tilesMap.get("f135").resize(SCALE);
    top.drawAt(im, "Left", startPt);
    corners[1] = top.getPtLoc("Head");
    corners[2] = top.getPtLoc("Tail");

    // tile up the left side of the triangle
    Tile left = tilesMap.get("f225").resize(SCALE);
    left.drawAt(im, "Left", startPt);
    corners[0] = left.getPtLoc("Tail");


    Tile reTile = new Tile(im, "rightEnd");
    for (int i=0; i < 3; i++)
      reTile.addPt("C"+i, corners[i]);

    return reTile;
  }   // end of makeRightEnd()



  private static Tile makeLeftEnd()
  // left end triangle of the fishes tiles
  {
    BufferedImage im = new BufferedImage(300, 300, 
                                 BufferedImage.TYPE_INT_ARGB);

    Point2D.Double startPt = new Point2D.Double(150, 150);
       // middle of the drawing area

    Point2D.Double[] corners = new Point2D.Double[3];

    // tile along the top of the triangle
    Tile top = tilesMap.get("f135").resize(SCALE);
    top.drawAt(im, "Left", startPt);
    corners[1] = top.getPtLoc("Head");
    corners[2] = top.getPtLoc("Tail");

    // tile down the right side of the triangle
    Tile right = tilesMap.get("f45").resize(SCALE);
    right.drawAt(im, "Left", startPt);
    corners[0] = right.getPtLoc("Head");


    Tile leTile = new Tile(im, "leftEnd");
    for (int i=0; i < 3; i++)
      leTile.addPt("C"+i, corners[i]);

    return leTile;
  }   // end of makeLeftEnd()




  private static Tile makeQuarter()
  {
    BufferedImage im = new BufferedImage(1000, 500, 
                                 BufferedImage.TYPE_INT_ARGB);  // alpha channel

    Point2D.Double tip = new Point2D.Double(500,500);
       // bottom edge, in the middle of the drawing area

    Tile cTile = tilesMap.get("f135").resize(1.0/SCALE);    // enlarge
    cTile.drawAt(im, "Left", tip);

    Point2D.Double startPt = cTile.getPtLoc("Tail");

    Pair<Point2D.Double, Point2D.Double> endPts = 
              drawLevels(1, 3, 1, im, startPt);    
         // returns top <left, right> corners of drawn quarter

    Tile quarTile = new Tile(im, "quarter");
    quarTile.addPt("C0", tip);
    quarTile.addPt("C1", endPts.getY());
    quarTile.addPt("C2", endPts.getX());

    return quarTile;
  }   // end of makeQuarter()



  private static Pair<Point2D.Double, Point2D.Double> 
            drawLevels(int level, int n, int numMids, 
                        BufferedImage im, Point2D.Double startPt)
  { if (level > n) {
      System.out.println("Level error");
      return null;
    }

    double lvlScale = Math.pow(0.5, level-1);

    // draw one row, left-to-right for level
    Tile t = tilesMap.get("leftEnd").resize(lvlScale); 
    t.drawAt(im, "C0",  startPt);
    Point2D.Double nextStartPt = t.getPtLoc("C2");
    Point2D.Double topLeft = t.getPtLoc("C1");

    for (int i=0; i < numMids; i++) {
      t = tilesMap.get("mid").resize(lvlScale); 
      t.drawAt(im, "C3", topLeft);
      topLeft = t.getPtLoc("C2");
    }

    t = tilesMap.get("rightEnd").resize(lvlScale); 
    t.drawAt(im, "C2",  topLeft);
    Point2D.Double nextEndPt = t.getPtLoc("C1");

    if (level == n)   // finished
      return new Pair<Point2D.Double, Point2D.Double>(nextStartPt, nextEndPt);
    else {
      return drawLevels(level+1, n, 2*numMids+2, im, nextStartPt); 
    }
  }  // end of drawLevels()



} // end of DrawFish class
