
// DrawMaze.java
// Andrew Davison, May 2022, ad@fivedots.coe.psu.ac.th

/* Draw a maze.

   Illustrates the use of many tiles, and the use of a
   complex set of point pairs to define how they can
   be combined. 

   Each tile is a square shape.
*/

import java.awt.geom.*;
import java.awt.image.*;
import java.util.*;


public class DrawMaze
{
  private static Tile noneTile, roundTile, turnTile,
                      pathTile, branchTile, crossTile;


  public static void main(String args[])
  {
    // -------- create the tiles -----------------------

    Point2D.Double[] pts = {  // each tile has four points; defined ccw
       new Point2D.Double(0,0), new Point2D.Double(0,60),
       new Point2D.Double(60,60), new Point2D.Double(60,0)  };

    // note: each tile uses a unique label
    noneTile = new Tile("data/none.png");
    noneTile.addPts("N", pts);

    roundTile = new Tile("data/round.png");
    roundTile.addPts("R", pts);

    turnTile = new Tile("data/turn.png");
    turnTile.addPts("T", pts);

    pathTile = new Tile("data/path.png");
    pathTile.addPts("P", pts);

    branchTile = new Tile("data/branch.png");
    branchTile.addPts("B", pts);

    crossTile = new Tile("data/cross.png");
    crossTile.addPts("C", pts);


    // ------- matching point pairs -------------

    /* Each tile side either has a maze path or is blank.

       "zero" means that no path enters/leaves a side; 
       "one" means there is a path entering/leaving a side
    */

    // ccw 'from' pts
    String[] zeroPts = {   // no path leaving the side
       "N0", "N1", "N2", "N3",  "R0", "R1", "R2",
       "T0", "T1",  "P0", "P2", "B0" };

    String[] onePts = {  // path leaving the side
       "R3",  "T2", "T3",  "P1", "P3", 
       "B1", "B2", "B3",  "C0", "C1", "C2", "C3" };

    // clockwise to pts
    String[] toZeros = {   // no path entering the side
       "N0", "N1", "N2", "N3",  
       "R1", "R2", "R3",  "T1", "T2",
       "P1", "P3",  "B1" };

    String[] toOnes = {  // path entering the side
       "R0", "T0", "T3",  "P0", "P2",
       "B0", "B2", "B3",  
       "C0", "C1", "C2", "C3" };

    // allow "no path" and path pairs
    PointPairs ptPairs = new PointPairs();
    ptPairs.add(zeroPts, toZeros);  // no paths
    ptPairs.add(onePts, toOnes);    // paths


    // draw tiles on the screen
    ImageViewer iview = Pics.view("Make a Maze", Pics.altScreen());
    Search.tileLocs(crossTile, ptPairs, iview, DrawMaze::genTile);
    Pics.save(iview.getImage(), "SavedPics/maze.png");
  } // end of main()



  private static Tile genTile(String ptName)
  // point names begin with 'N', 'R', 'T', 'P', 'B', or 'C'
  {
    char ch = ptName.charAt(0);
    switch(ch) {
      case 'N': return noneTile.clone();
      case 'R': return roundTile.clone();
      case 'T': return turnTile.clone();
      case 'P': return pathTile.clone();
      case 'B': return branchTile.clone();
      case 'C': return crossTile.clone();

      default: System.out.println("Unknown point name: " + ptName);
    }
    return null;
  }  // end of genTile()



} // end of DrawMaze class
