
// DrawPenrose.java
// Andrew Davison, May 2022, ad@fivedots.coe.psu.ac.th

/* Tiling using Penrose's Kite and dart tiling (P2)
   e.g. see https://en.wikipedia.org/wiki/Penrose_tiling
            https://mathworld.wolfram.com/PenroseTiles.html
   Martin Gardner's article:
     https://www.maa.org/sites/default/files/pdf/pubs/focus/Gardner_PenroseTilings1-1977.pdf
   Also see PenroseClusters.java

*/

import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.util.*;
import java.util.function.*;



public class DrawPenrose
{
  private static Tile dart, kite;


  public static void main(String args[])
  {
    dart = buildDart();
    kite = buildKite();

    dart.view();
    dart.reportPoints();
    kite.view();
    kite.reportPoints();

    PointPairs ptPairs = new PointPairs();
    ptPairs.add("D0", "D0");  // dart pts
    ptPairs.add("D0", "K0");
    ptPairs.add("D1", "K1");
    ptPairs.add("D2", "K2");
    ptPairs.add("D3", "K3");
    ptPairs.add("D3", "D1");
			 
    ptPairs.add("K0", "D2");  // kite pts
    ptPairs.add("K0", "K2");
    ptPairs.add("K1", "D3");
    ptPairs.add("K1", "K1");
    ptPairs.add("K2", "D0");
    ptPairs.add("K2", "K0");
    ptPairs.add("K3", "D1");
    ptPairs.add("K3", "K3");

    // PointPairs ptPairs = PointPairs.matchLens( new Tile[]{dart,kite});
      /* Does not work due to concave sides (and too general).
         It generates:
             D0 -> [D0, D1, K0, K3]
             D1 -> [D2, D3, K1, K2]
             D2 -> [D2, D3, K1, K2]
             D3 -> [D0, D1, K0, K3]
             K0 -> [D2, D3, K1, K2]
             K1 -> [D2, D3, K1, K2]
             K2 -> [D0, D1, K0, K3]
             K3 -> [D0, D1, K0, K3]
      */


    // draw tiles on the screen
    ImageViewer iview =  Pics.view("Penrose Tiles", Pics.altScreen());  

    // Search.tileSpacey(dart, ptPairs, iview, DrawPenrose::genTile);
    Search.tileLocs(dart, ptPairs, iview, DrawPenrose::genTile);
      /* tileLocs() is better than tileSpacey() but still makes poor
         tiling choices that leave gaps. A more complex backtracking
         search is needed! Also, see Isoceles.java
      */

    Pics.save(iview.getImage(), "SavedPics/penrose.png");
  } // end of main()




  private static Tile genTile(String ptName)
  // point names begin with 'D' or 'K'
  {
    return (ptName.charAt(0) == 'D') ? dart.clone() : kite.clone();
  }


  // --------------------- build tiles --------------------------
  // this code is also used in PenroseClusters.java

  private static Tile buildDart()
  {
    Tile t = new Tile("data/dartTile.txt");
    /* using angles/lengths from 
       http://jwilson.coe.uga.edu/emat6680fa05/schultz/penrose/penrose_main.html
    */
    ArrayList<Point2D.Double> pts = t.getImPts();

    // add red circles at two corners
    Graphics2D g2d = t.getGraphics();
    g2d.setColor(Color.RED);
    g2d.fillOval((int)pts.get(1).x-22, (int)pts.get(1).y-9, 
                                     6, 6);   // below/left of pt1

    g2d.fillOval((int)pts.get(3).x-3, (int)pts.get(3).y+19, 
                                     6, 6);   // above/right of pt3

    double[] dxs = { 50, 52, 78, 60, 36, 50, 95 };
    double[] dys = { 32, 68, 86, 88, 86, 44, 89 };
    t.addInners(dxs, dys);

    t.addOuter("D0", "D1", 67, 107);
    t.addOuter("D1", "D2", 99, 73);
    t.addOuter("D2", "D3", 71, 30);
    t.addOuter("D3", "D0", 27, 48);

    return t;
  }  // end of buildDart()



  private static Tile buildKite()
  {
    Tile t = new Tile("data/kiteTile.txt");
    /* using angles/lengths from 
       http://jwilson.coe.uga.edu/emat6680fa05/schultz/penrose/penrose_main.html
    */

    ArrayList<Point2D.Double> pts = t.getImPts();

    // add red circles at two corners
    Graphics2D g2d = t.getGraphics();
    g2d.setColor(Color.RED);

    g2d.fillOval((int)pts.get(1).x-6, (int)pts.get(1).y-19, 
                                      6, 6);   // below/left of pt1

    g2d.fillOval((int)pts.get(3).x+1, (int)pts.get(3).y+17, 
                                      6, 6);   // above/right of pt3

    double[] dxs = { 58, 79,107, 77, 47, 46, 46, 97, 35, 29, 62 };
    double[] dys = { 45, 60, 77,104,109, 79, 62, 91, 97,109,107 };
    t.addInners(dxs, dys);

    t.addOuter("K0", "K1",  50, 126);
    t.addOuter("K1", "K2", 111, 107);
    t.addOuter("K2", "K3", 93, 44);
    t.addOuter("K3", "K0", 23, 71);

    return t;
  }  // end of buildKite()


} // end of DrawPenrose class
