
// Peanuts.java
// Andrew Davison, May 2022, ad@fivedots.coe.psu.ac.th

/* More complex row-by-row tiling based around Tile.testDrawAt()
   which may succeed or fail to draw a tile.

   Internally Tile.testDrawAt() uses inner and outer points to decide
   if the tile can be drawn. In this example, they are generated
   automatically, and can be seen by calling Tile.view().
*/

import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.util.*;



public class Peanuts
{
  private static final int NUM_NUTS = 6;

  private static Random rand = new Random();



  public static void main(String args[])
  {
    // A peanut with a blue dot at one end
    Tile p = new Tile("data/nutTile.txt");
    Graphics2D g2d = p.getGraphics();
    g2d.setColor(Color.BLUE);
    g2d.fillOval(110, 36, 10, 10);  // coord obtained using ViewTile
    p.view();     // shows the inner and outer points
    p.reportPoints();

    // create tiles for all the possible rotations of the peanut
    Tile[] peanuts = new Tile[NUM_NUTS];
    for (int i=0; i < NUM_NUTS; i++) {
      peanuts[i] = p.rotate(60*i);
      peanuts[i].view();    // shows the inner and outer points
    }

    String[] drawPts = new String[] {
          "N6", "N5", "N4", "N3", "N2", "N7" };
     /* each tile should be drawn using a different point,
        which are specified in this array */

    tilePeanuts(peanuts, drawPts);
  } // end of main()




  private static void tilePeanuts(Tile[] peanuts, String[] drawPts)
  // draw the peanuts row-by-row
  { 
    // calculate offsets between the tiles
    Point2D.Double pt6 = peanuts[0].getCartesian("N6");
    Point2D.Double pt2 = peanuts[0].getCartesian("N2");
    Point2D.Double pt4 = peanuts[0].getCartesian("N4");
    double xOffset = Math.abs(pt4.x - pt6.x);
    double yOffset = Math.abs(pt2.y - pt6.y)/2.0;
    // System.out.printf("Offsets: (%.2f, %.2f)\n", xOffset, yOffset);

    ImageViewer iview = Pics.view("Peanuts", Pics.altScreen());
    BufferedImage scrImage = iview.getImage();
    int pWidth = scrImage.getWidth();
    int pHeight = scrImage.getHeight();

    int rowNum = 0;
    double y = 0;
    while (y < pHeight) {
      double x = (rowNum%2 == 0)? 10 : 10+xOffset;
      while (x < pWidth) {
        if (Pics.isTransparent(scrImage, x, y+yOffset)) {
          boolean isDrawn = 
              randPeanut(peanuts, drawPts, x, y, scrImage);
          if (isDrawn)
            iview.repaint();
        }
        x += 2*xOffset;
      }
      y += yOffset;
      rowNum++;
    }

    Pics.save(scrImage, "SavedPics/peanuts.png");
  }  // end of tilePeanuts()


  private static boolean randPeanut(
                      Tile[] peanuts, String[] drawPts,  
                      double x, double y, BufferedImage scrImage)
  /* Loop through the peanut tiles until one of them is
     successfully drawn. Note that Tile.testDrawAt() uses automatically
     generated inner and outer points since I don't create any myself.

     The looping is randomized by generating a random sequence of
     indicies between 0 - (NUM_NUTS-1) (e.g. {3, 5, 1, 0, 2, 4})
  */
  {
    int[] randIdxs = shuffleSeq(NUM_NUTS);  // e.g. {3, 5, 1, 0, 2, 4}

    for (int i=0; i < NUM_NUTS; i++) {
      int idx = randIdxs[i];
      if ( peanuts[idx].testDrawAt(scrImage, drawPts[idx], x, y))
        return true;
    }
    // System.out.println("No suitable tile found");
    return false;
  }


  private static int[] shuffleSeq(int n)
  // a randomly ordered sequence of ints 0..(n-1)
  {
    int[] ids = new int[n];
    for (int i=0; i < n; i++)
      ids[i] = i;

    for (int i=0; i < n; i++)
      swap(ids, i, rand.nextInt(n));

    return ids;
  }

  private static void swap(int[] a, int i, int j)
  {
    int temp = a[i];
    a[i] = a[j];
    a[j] = temp;
  }


} // end of Peanuts class
