
// PenroseClusters.java
// Andrew Davison, May 2022, ad@fivedots.coe.psu.ac.th

/* Tiling using Penrose's Kite and dart tiling (P2)
   e.g. see https://en.wikipedia.org/wiki/Penrose_tiling
            https://mathworld.wolfram.com/PenroseTiles.html

   See DrawPenrose.java for actual tiling code.

   This program illustrates the use of composition to make
   the Penrose P2 clusters (e.g. see
     https://commons.wikimedia.org/wiki/File:Penrose_vertex_figures.png
   and Martin Gardner's article:
    https://www.maa.org/sites/default/files/pdf/pubs/focus/Gardner_PenroseTilings1-1977.pdf
*/

import java.awt.*;
import java.awt.geom.*;
import java.awt.image.*;
import java.util.*;



public class PenroseClusters
{
  private static final int TILE_WIDTH = 160;     
  private static final int TILE_HEIGHT = 160;

  private static Tile dart, kite;



  public static void main(String args[])
  {
    dart = buildDart();
    dart.view();
    dart.reportPoints();

    kite = buildKite();
    kite.view();
    kite.reportPoints();

    Tile[] cluster = new Tile[7];
    cluster[0] = makeAceTile();
    cluster[1] = makeDeuceTile();
    cluster[2] = makeJackTile();
    cluster[3] = makeKingTile();
    cluster[4] = makeQueenTile();
    cluster[5] = makeStarTile();
    cluster[6] = makeSunTile();

    for(Tile t : cluster) {
      t.view();
      t.reportPoints();
    }
  } // end of main()



  // --------------------- build tiles --------------------------
  // same code as in DrawPenrose.java


  private static Tile buildDart()
  {
    Tile t = new Tile("data/dartTile.txt");
    /* using angles/lengths from 
       http://jwilson.coe.uga.edu/emat6680fa05/schultz/penrose/penrose_main.html
    */

    ArrayList<Point2D.Double> pts = t.getImPts();

    // add red circles at two corners
    Graphics2D g2d = t.getGraphics();
    g2d.setColor(Color.RED);
    g2d.fillOval((int)pts.get(1).x-22, (int)pts.get(1).y-9, 
                                     6, 6);   // below/left of pt1

    g2d.fillOval((int)pts.get(3).x-3, (int)pts.get(3).y+19, 
                                     6, 6);   // above/right of pt3

    double[] dxs = { 50, 52, 78, 60, 36, 50, 95 };
    double[] dys = { 32, 68, 86, 88, 86, 44, 89 };
    t.addInners(dxs, dys);

    t.addOuter("D0", "D1", 67, 107);
    t.addOuter("D1", "D2", 99, 73);
    t.addOuter("D2", "D3", 71, 30);
    t.addOuter("D3", "D0", 27, 48);

    return t;
  }  // end of buildDart()



  private static Tile buildKite()
  {
    Tile t = new Tile("data/kiteTile.txt");
    /* using angles/lengths from 
       http://jwilson.coe.uga.edu/emat6680fa05/schultz/penrose/penrose_main.html
    */

    ArrayList<Point2D.Double> pts = t.getImPts();

    // add red circles at two corners
    Graphics2D g2d = t.getGraphics();
    g2d.setColor(Color.RED);

    g2d.fillOval((int)pts.get(1).x-6, (int)pts.get(1).y-19, 
                                      6, 6);   // below/left of pt1

    g2d.fillOval((int)pts.get(3).x+1, (int)pts.get(3).y+17, 
                                      6, 6);   // above/right of pt3

    double[] dxs = { 58, 79,107, 77, 47, 46, 46, 97 };
    double[] dys = { 45, 60, 77,104,109, 79, 62, 91 };
    t.addInners(dxs, dys);

    t.addOuter("K0", "K1",  50, 126);
    t.addOuter("K1", "K2", 111, 107);
    t.addOuter("K2", "K3", 93, 44);
    t.addOuter("K3", "K0", 23, 71);

    return t;
  }  // end of buildKite()


  // -------------- build clusters ------------------------------


  private static Tile makeAceTile()
  // generate 'Ace' Penrose cluster
  {
    int w = TILE_WIDTH*2; 
    int h = TILE_HEIGHT*2;
    BufferedImage im = new BufferedImage(w, h, 
                                 BufferedImage.TYPE_INT_ARGB);  // alpha channel

    int numCorners = 4;
    Point2D.Double[] corners = new Point2D.Double[numCorners];

    // Draw an 'Ace' Penrose image around center of image, and collect corners
    Tile leftKite = kite.rotate(18);
    leftKite.drawAt(im, "K2", w/2, h/2);
    corners[0] = leftKite.getPtLoc("K3");   // will become A0
    corners[1] = leftKite.getPtLoc("K0");   // A1

    Tile rightKite = kite.rotate(90);
    rightKite.drawAt(im, "K0", w/2, h/2);
    corners[3] = rightKite.getPtLoc("K2");   // A3
    
    Tile bottomDart = dart.rotate(36+90);
    bottomDart.drawAt(im, "D2", w/2, h/2);
    corners[2] = bottomDart.getPtLoc("D0");   // A2
    
    Pics.save(im, "SavedPics/ace.png");
    // System.out.println("Corners array: " + Arrays.toString(corners));

    // create a tile using the image and corners
    Tile t = new Tile(im, "ace");
    for (int i=0; i < numCorners; i++)
      t.addPt("A"+i, corners[i]);

    return t;
  }   // end of makeAceTile()



  private static Tile makeDeuceTile()
  // generate 'Deuce' Penrose cluster
  {
    int w = TILE_WIDTH*2; 
    int h = TILE_HEIGHT*2;
    BufferedImage im = new BufferedImage(w, h, 
                                 BufferedImage.TYPE_INT_ARGB);  // alpha channel
    // Graphics2D g2d = Pics.createGraphics(im);

    int numCorners = 8;
    Point2D.Double[] corners = new Point2D.Double[numCorners];

    // Draw a 'Deuce' Penrose image around center of image, and collect corners
    Tile leftDart = dart.rotate(-90);
    leftDart.drawAt(im, "D3", w/2, h/2);
    corners[0] = leftDart.getPtLoc("D0");   // will become De0
    corners[1] = leftDart.getPtLoc("D1");   // De1
    corners[2] = leftDart.getPtLoc("D2");   // De2

    Tile rightDart = dart.rotate(-18);
    rightDart.drawAt(im, "D1", w/2, h/2);
    corners[7] = rightDart.getPtLoc("D3");   // De7
    corners[6] = rightDart.getPtLoc("D2");   // De6
    
    Tile leftKite = kite.rotate(180-18);
    leftKite.drawAt(im, "K1", w/2, h/2);
    corners[3] = leftKite.getPtLoc("K3");   // De3
    corners[4] = leftKite.getPtLoc("K0");   // De4
    
    Tile rightKite = kite.rotate(36-90);
    rightKite.drawAt(im, "K1", w/2, h/2);
    corners[5] = rightKite.getPtLoc("K3");   // De5
    

    Pics.save(im, "SavedPics/deuce.png");
    // System.out.println("Corners array: " + Arrays.toString(corners));

    // create a tile using the image and corners
    Tile t = new Tile(im, "deuce");
    for (int i=0; i < numCorners; i++)
      t.addPt("De"+i, corners[i]);

    return t;
  }   // end of makeDeuceTile()



  private static Tile makeJackTile()
  // generate 'Jack' Penrose cluster
  {
    int w = TILE_WIDTH*2; 
    int h = TILE_HEIGHT*2;
    BufferedImage im = new BufferedImage(w, h, 
                                 BufferedImage.TYPE_INT_ARGB);  // alpha channel
    // Graphics2D g2d = Pics.createGraphics(im);

    int numCorners = 10;
    Point2D.Double[] corners = new Point2D.Double[numCorners];

    // Draw a 'Jack' Penrose image around center of image, and collect corners
    Tile leftKite = kite.rotate(-90);
    leftKite.drawAt(im, "K3", w/2, h/2);
    corners[0] = leftKite.getPtLoc("K0");   // J0
    corners[1] = leftKite.getPtLoc("K1");   // J1
    corners[2] = leftKite.getPtLoc("K2");   // J2

    Tile rightKite = kite.rotate(-72-90);
    rightKite.drawAt(im, "K3", w/2, h/2);
    corners[8] = rightKite.getPtLoc("K0");   // J8
    corners[9] = rightKite.getPtLoc("K1");   // J9

    Tile leftDart = dart.rotate(-90+72);
    leftDart.drawAt(im, "D3", w/2, h/2);
    corners[3] = leftDart.getPtLoc("D1");   // J3
    corners[4] = leftDart.getPtLoc("D2");   // J4

    Tile rightDart = dart.rotate(-18-72);
    rightDart.drawAt(im, "D1", w/2, h/2);
    corners[7] = rightDart.getPtLoc("D3");   // J7
    corners[6] = rightDart.getPtLoc("D2");   // J6
    
    Tile bottomKite = kite.rotate(-36-90);
    bottomKite.drawAt(im, "K1", w/2, h/2);
    corners[5] = bottomKite.getPtLoc("K3");   // J5
    
    Pics.save(im, "SavedPics/jack.png");

    // create a tile using the image and corners
    Tile t = new Tile(im, "jack");
    for (int i=0; i < numCorners; i++)
      t.addPt("J"+i, corners[i]);

    return t;
  }   // end of makeJackTile()



  private static Tile makeKingTile()
  // generate 'King' Penrose cluster
  {
    int w = TILE_WIDTH*2; 
    int h = TILE_HEIGHT*2;
    BufferedImage im = new BufferedImage(w, h, 
                                 BufferedImage.TYPE_INT_ARGB);  // alpha channel
    // Graphics2D g2d = Pics.createGraphics(im);

    int numCorners = 10;
    Point2D.Double[] corners = new Point2D.Double[numCorners];

    // Draw a 'King' Penrose image around center of image, and collect corners
    Tile leftDart = dart.rotate(90+36+72);
    leftDart.drawAt(im, "D0", w/2, h/2);
    corners[0] = leftDart.getPtLoc("D1");   // will become Ki0
    corners[1] = leftDart.getPtLoc("D2");   // Ki1
    corners[2] = leftDart.getPtLoc("D3");   // Ki2

    Tile midDart = dart.rotate(90+36);
    midDart.drawAt(im, "D0", w/2, h/2);
    corners[9] = midDart.getPtLoc("D2");   // Ki9
    corners[8] = midDart.getPtLoc("D1");   // Ki8
    
    Tile rightDart = dart.rotate(90-36);
    rightDart.drawAt(im, "D0", w/2, h/2);
    corners[7] = rightDart.getPtLoc("D2");   // Ki7
    corners[6] = rightDart.getPtLoc("D1");   // Ki6
    
    Tile leftKite = kite.rotate(36+90);
    leftKite.drawAt(im, "K2", w/2, h/2);
    corners[4] = leftKite.getPtLoc("K1");   // Ki4
    corners[3] = leftKite.getPtLoc("K0");   // Ki3
    
    Tile rightKite = kite.rotate(-18);
    rightKite.drawAt(im, "K0", w/2, h/2);
    corners[5] = rightKite.getPtLoc("K2");   // Ki5
    
    Pics.save(im, "SavedPics/king.png");

    // create a tile using the image and corners
    Tile t = new Tile(im, "king");
    for (int i=0; i < numCorners; i++)
      t.addPt("Ki"+i, corners[i]);

    return t;
  }   // end of makeKingTile()



  private static Tile makeQueenTile()
  // generate 'Queen' Penrose cluster
  {
    int w = TILE_WIDTH*2; 
    int h = TILE_HEIGHT*2;
    BufferedImage im = new BufferedImage(w, h, 
                                 BufferedImage.TYPE_INT_ARGB);  // alpha channel
    // Graphics2D g2d = Pics.createGraphics(im);

    int numCorners = 10;
    Point2D.Double[] corners = new Point2D.Double[numCorners];

    // Draw a 'Queen' Penrose image around center of image, and collect corners
    Tile leftKite = kite.rotate(-36+90);
    leftKite.drawAt(im, "K2", w/2, h/2);
    corners[0] = leftKite.getPtLoc("K3");   // Q0
    corners[1] = leftKite.getPtLoc("K0");   // Q1
    corners[2] = leftKite.getPtLoc("K1");   // Q2

    Tile midDart = dart.rotate(90+36);
    midDart.drawAt(im, "D0", w/2, h/2);
    corners[9] = midDart.getPtLoc("D2");   // Q9
    corners[8] = midDart.getPtLoc("D1");   // Q8

    Tile rightKite = kite.rotate(-36+90);
    rightKite.drawAt(im, "K0", w/2, h/2);
    corners[7] = rightKite.getPtLoc("K2");   // Q7
    corners[6] = rightKite.getPtLoc("K1");   // Q6

    Tile leftBottomKite = kite.rotate(-90);
    leftBottomKite.drawAt(im, "K0", w/2, h/2);
    corners[3] = leftBottomKite.getPtLoc("K2");   // Q3
    corners[4] = leftBottomKite.getPtLoc("K3");   // Q4
    
    Tile rightBottomKite = kite.rotate(108+90);
    rightBottomKite.drawAt(im, "K2", w/2, h/2);
    corners[5] = rightBottomKite.getPtLoc("K0");   // Q5
    
    Pics.save(im, "SavedPics/queen.png");

    // create a tile using the image and corners
    Tile t = new Tile(im, "queen");
    for (int i=0; i < numCorners; i++)
      t.addPt("Q"+i, corners[i]);

    return t;
  }   // end of makeQueenTile()



  private static Tile makeStarTile()
  // generate 'Star' Penrose cluster
  {
    int w = TILE_WIDTH*2; 
    int h = TILE_HEIGHT*2;
    BufferedImage im = new BufferedImage(w, h, 
                                 BufferedImage.TYPE_INT_ARGB);  // alpha channel
    // Graphics2D g2d = Pics.createGraphics(im);

    int numCorners = 10;
    Point2D.Double[] corners = new Point2D.Double[numCorners];

    // Draw a 'Star' Penrose image around center of image, and collect corners

    double startAngle = 72+90;
    for (int i=0; i < 5; i++) {
      Tile d = dart.rotate(startAngle + 72*i);
      d.drawAt(im, "D0", w/2, h/2);
      corners[2*i] = d.getPtLoc("D1");   // S (2*i)
      corners[(2*i)+1] = d.getPtLoc("D2");   // S (2*i+1)
    }
    
    Pics.save(im, "SavedPics/star.png");

    // create a tile using the image and corners
    Tile t = new Tile(im, "star");
    for (int i=0; i < numCorners; i++)
      t.addPt("S"+i, corners[i]);

    return t;
  }   // end of makeStarTile()



  private static Tile makeSunTile()
  // generate 'Sun' Penrose cluster
  {
    int w = TILE_WIDTH*2; 
    int h = TILE_HEIGHT*2;
    BufferedImage im = new BufferedImage(w, h, 
                                 BufferedImage.TYPE_INT_ARGB);  // alpha channel
    // Graphics2D g2d = Pics.createGraphics(im);

    int numCorners = 10;
    Point2D.Double[] corners = new Point2D.Double[numCorners];

    // Draw a 'Sun' Penrose image around center of image, and collect corners

    double startAngle = -36-90;
    for (int i=0; i < 5; i++) {
      Tile k = kite.rotate(startAngle + 72*i);
      k.drawAt(im, "K3", w/2, h/2);
      corners[2*i] = k.getPtLoc("K1");   // Su (2*i)
      corners[(2*i)+1] = k.getPtLoc("K2");   // Su (2*i+1)
    }
    
    Pics.save(im, "SavedPics/sun.png");

    // create a tile using the image and corners
    Tile t = new Tile(im, "sun");
    for (int i=0; i < numCorners; i++)
      t.addPt("Su"+i, corners[i]);

    return t;
  }   // end of makeSunTile()


} // end of PenroseClusters class
